from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.shortcuts import *

__all__ = ["compute_telluric_template_alternative", "plot_telluric_template_alternative"]

def compute_telluric_template_alternative(config_in):

    compute_telluric_template_alternative_routine(config_in,
                                n_iterations=1,
                                use_berv=False,
                                use_reference_airmass=False,
                                use_template=True,
                                subroutine_name='telluric_template')


def compute_telluric_template_alternative_routine(config_in, **kwargs):
    """
    Lazy workaround
    :param config_in:
    :param kwargs:
    :return:
    """

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)

    shared_data = load_from_cpickle('shared', config_in['output'])

    for night in night_dict:

        instrument_name = night_dict[night]['instrument']

        print()
        print("compute_telluric_template                  Night: ", night)
        print()

        try:
            telluric = load_from_cpickle('telluric', config_in['output'], night)
            continue
        except:
            print("  No telluric correction file found, computing now ")
            print()



        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        """ Retrieving the observations"""
        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'], use_telluric=False)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        processed = {
            'subroutine': kwargs['subroutine_name'],
            'n_orders': 0,
            'n_pixels': 0,
            'telluric': {}
        }

        telluric = {
            'subroutine': kwargs['subroutine_name'],
            'reference_frame': 'observer',
            'template': {},
            'linear' : {}
        }

        # There must be a more elegant way to do this, but I'm, not aware of it
        """ computation of the rescaled spectra, for later use in the analysis and in the plotting subroutines"""
        for obs in lists['observations']:
            processed[obs] = {
                'n_orders': input_data[obs]['n_orders'],
                'n_pixels': input_data[obs]['n_pixels']
            }

            """ for plotting purpose only"""
            processed[obs]['wave'] = input_data[obs]['wave']
            processed[obs]['e2ds'] = input_data[obs]['e2ds']
            processed[obs]['e2ds_err'] = input_data[obs]['e2ds_err']

            processed[obs]['e2ds_rescaling'], processed[obs]['e2ds_rescaled'], processed[obs]['e2ds_rescaled_err'] = \
                perform_rescaling(input_data[obs]['wave'],
                                  input_data[obs]['e2ds'],
                                  input_data[obs]['e2ds_err'],
                                  observational_pams['wavelength_rescaling'])

        """ Reference airmass for iterative correction of airmass - disabled here"""
        processed['airmass_ref'] = 0.000

        """ 
        Definiton of the wavelength scale for the output telluric spectrum
        We assume that the wavelength solution did not change during the night 
        """
        obs_reference = lists['observations'][0]
        telluric['rebinned'] = {
                'wave': input_data[obs_reference]['wave'],
                'step': input_data[obs_reference]['step'],
                'flux': np.ones(input_data[obs_reference]['step'].shape),
                'ferr': np.ones(input_data[obs_reference]['step'].shape)*0.001
        }

        processed['telluric']['spectrum_noairmass'] = telluric['rebinned']['flux'].copy()
        processed['telluric']['spectrum_noairmass_err'] = telluric['rebinned']['ferr'].copy()
        processed['n_orders'] = input_data[obs_reference]['orders']
        processed['n_pixels'] = input_data[obs_reference]['wave_size']

        """ 
        Computation of the rescaling factor for the telluric template. This factor has no physical value, it's just 
        the ratio of the observed telluric features (in a normalized spectra) to the input template.
        """
        try:

            if 'telluric_template' not in instrument_dict[instrument_name]:
                raise MissingKeywordException()

            template_dict = instrument_dict[instrument_name]['telluric_template']

            if template_dict['fit_range'][0] > shared_data['coadd']['wavelength_range'][1] or \
                    template_dict['fit_range'][1] < shared_data['coadd']['wavelength_range'][0]:
                raise OutOfRangeException()

            print('    rescaled telluric template')
            print('    instrument :', instrument_name)
            print('    template   :', template_dict['file'])
            print('    fit_range  :', template_dict['fit_range'])
            print()

            """ Retrieving the template data"""
            telluric_template_data = np.genfromtxt(template_dict['file'])

            telluric['template']['input'] = {
                'range': [np.amin(telluric_template_data[:, 0]), np.amax(telluric_template_data[-1, 0])],
                'wave': telluric_template_data[:, 0],
                'flux': telluric_template_data[:, 1],
                'ferr': telluric_template_data[:, 2],
                'step': telluric_template_data[:, 3]
            }

            processed['template'] = {
                'array_wave': [],
                'array_data': {},
                'array_move': {}
            }

            """ Again we assume that the wavelength solution did not change during the night  """

            tel_selection = (input_data[obs_reference]['wave'] > template_dict['fit_range'][0]) & \
                            (input_data[obs_reference]['wave'] < template_dict['fit_range'][1])
            processed['template']['wave_sel'] = input_data[obs_reference]['wave'][tel_selection]
            processed['template']['step_sel'] = input_data[obs_reference]['step'][tel_selection]

            for obs in lists['telluric']:
                processed['template'][obs] = processed[obs]['e2ds_rescaled'][tel_selection]
                """ saving the wavelength array for plotting purpose"""
                processed['template']['array_wave'].extend(processed['template']['wave_sel'])

            rescaling_array = np.arange(0.05, 2.0, 0.05)
            computed_std = np.zeros(len(rescaling_array))

            for n_rescaling_factor, rescaling_factor in enumerate(rescaling_array):

                """ 
                Template spectrum is rebinned onto the observations wavelength scale, using only the wavelength 
                range selected for the computation of the rescaling factor 
                """
                template_rebinned_flux = \
                    rebin_1d_to_1d(telluric['template']['input']['wave'],
                                   telluric['template']['input']['step'],
                                   telluric['template']['input']['flux'],
                                   processed['template']['wave_sel'],
                                   processed['template']['step_sel'],
                                   preserve_flux=False)

                """ Saving the outcome to dictionary """
                template_rebinned_flux -= 1.00
                template_rebinned_flux *= rescaling_factor
                template_rebinned_flux += 1.00

                e2ds_corrected = []

                for obs in lists['telluric']:

                    e2ds_corrected.extend(processed['template'][obs] /
                                            np.power(template_rebinned_flux, input_data[obs]['AIRMASS']))

                computed_std[n_rescaling_factor] = np.nanstd(e2ds_corrected)

                #print(n_rescaling_factor, rescaling_factor, computed_std[n_rescaling_factor])
                #plt.scatter(processed['template']['array_wave'], e2ds_corrected)
                #plt.plot(processed['template']['wave_sel'], template_rebinned_flux)
                #plt.show()


                label_dict = '{0}'.format(rescaling_factor)
                processed['template']['array_data'][label_dict] = e2ds_corrected
                processed['template']['array_move'][label_dict] = rescaling_factor


            #plt.scatter(rescaling_array, computed_std)
            #plt.show()

            """ Selection of the rescaling factor with the lowest scatter """
            ind_factor = np.argmin(computed_std)
            ind_range = 3
            if ind_factor < ind_range:
                sel_factor = rescaling_array[0:ind_factor+ind_range]
                sel_stdev = computed_std[0:ind_factor+ind_range]
            elif ind_factor > len(rescaling_array) - ind_range:
                sel_factor = rescaling_array[ind_factor-ind_range:]
                sel_stdev = computed_std[ind_factor-ind_range:]
            else:
                sel_factor = rescaling_array[ind_factor-ind_range:ind_factor+ind_range]
                sel_stdev = computed_std[ind_factor-ind_range:ind_factor+ind_range]

            coeff = np.polyfit(sel_factor, sel_stdev, 2)

            telluric_factor = - coeff[1] / (2*coeff[0])

            print('   telluric factor: {0:7f}'.format(telluric_factor))
            print()

            processed['template']['telluric_factor'] = telluric_factor
            processed['template']['rescaling_factor'] = sel_factor
            processed['template']['computed_std'] = computed_std
            processed['template']['polyfit'] = {
                'package': 'numpy', # in case we forget what we used...
                'order': 2,
                'coeff': coeff,
                'sel_factor': sel_factor,
                'sel_stdev': sel_stdev
             }

            """ 
            The template telluric spectrum is rebinned onto the 2D scale of the observations.
            Then it is rescaled according to the computed factor
            We assume that the wavelength solution did not change during the night
            """

            processed['template']['rebinned'] = {}
            processed['template']['rebinned']['flux'] = \
                rebin_1d_to_2d(telluric['template']['input']['wave'],
                               telluric['template']['input']['step'],
                               telluric['template']['input']['flux'],
                               telluric['rebinned']['wave'],
                               telluric['rebinned']['step'],
                               preserve_flux=False)

            processed['template']['rebinned']['ferr'] = \
                rebin_1d_to_2d(telluric['template']['input']['wave'],
                               telluric['template']['input']['step'],
                               telluric['template']['input']['ferr'],
                               telluric['rebinned']['wave'],
                               telluric['rebinned']['step'],
                               preserve_flux=False,
                               is_error=True)

            sel_out_of_range = ~((telluric['rebinned']['wave'] > telluric['template']['input']['range'][0]+1.) \
                                & (telluric['rebinned']['wave'] < telluric['template']['input']['range'][1]-1.))
            processed['template']['rebinned']['flux'][sel_out_of_range] = 1.
            processed['template']['rebinned']['ferr'][sel_out_of_range] = 0.1

            processed['telluric']['spectrum_noairmass'] = \
                (processed['template']['rebinned']['flux'] - 1.) * telluric_factor + 1.0

            processed['telluric']['spectrum_noairmass_err'] = processed['template']['rebinned']['ferr'] * telluric_factor

        except MissingFileException:
            print(' *** Missing telluric_template keyword in configuration file ***')
            print()

        except OutOfRangeException:
            print(' *** Wavelength range for the calculation of the rescaling factor is outside the boundaries ***')
            print('     Rescaling factor wavelength range: {0:7.2f} to {1:7.2f}'.format(
                template_dict['fit_range'][0], template_dict['fit_range'][1]))
            print('     Shared data wavelength range     : {0:7.2f} to {1:7.2f}'.format(
                shared_data['coadd']['wavelength_range'][0],
                shared_data['coadd']['wavelength_range'][1]))

            print()

        """
        Introduction of a second telluric correction, where the telluric spectrum depends linearly on the precipitable 
        water vapour (PWV). As such, the values stored in the files are the coefficient of the PWV term, while the 
        baseline is given by the outcome of the previous step, i.e., the spectrum_noairmass array
        """
        try:

            """ 
            The algorith is essentially the same as in the previous step, with the exception of the calculation
            of the telluric spectrum at each iteration of the chi-square minimization
            """

            if 'telluric_linear_term' not in instrument_dict[instrument_name]:
                raise MissingKeywordException()

            linear_dict = instrument_dict[instrument_name]['telluric_linear_term']

            if linear_dict['fit_range'][0] > shared_data['coadd']['wavelength_range'][1] or \
                    linear_dict['fit_range'][1] < shared_data['coadd']['wavelength_range'][0]:
                raise OutOfRangeException()

            print('    PWV-dependent telluric spectrum')
            print('    instrument :', instrument_name)
            print('    linear     :', linear_dict['file'])
            print('    fit_range  :', linear_dict['fit_range'])
            print()

            """ Retrieving the linear data"""
            telluric_linear_data = np.genfromtxt(linear_dict['file'])

            telluric['linear']['input'] = {
                'range': [np.amin(telluric_linear_data[:, 0]), np.amax(telluric_linear_data[-1, 0])],
                'wave': telluric_linear_data[:, 0],
                'coef': telluric_linear_data[:, 1],
                'cerr': telluric_linear_data[:, 2],
                'step': telluric_linear_data[:, 3]
            }

            processed['linear'] = {
                'array_wave': [],
                'array_data': {},
                'array_move': {}
            }

            """ Again we assume that the wavelength solution did not change during the night  """
            tel_selection = (input_data[obs_reference]['wave'] > linear_dict['fit_range'][0]) & \
                            (input_data[obs_reference]['wave'] < linear_dict['fit_range'][1])
            processed['linear']['wave_sel'] = input_data[obs_reference]['wave'][tel_selection]
            processed['linear']['step_sel'] = input_data[obs_reference]['step'][tel_selection]

            for obs in lists['telluric']:
                processed['linear'][obs] = processed[obs]['e2ds_rescaled'][tel_selection]
                """ saving the wavelength array for plotting purpose"""
                processed['linear']['array_wave'].extend(processed['linear']['wave_sel'])

            rescaling_array = 10**np.arange(-1, np.log10(50), 0.1)
            computed_std = np.zeros(len(rescaling_array))

            for n_rescaling_factor, rescaling_factor in enumerate(rescaling_array):

                """ 
                Template spectrum is rebinned onto the observations wavelength scale, using only the wavelength 
                range selected for the computation of the rescaling factor 
                """
                linear_rebinned_flux = \
                    rebin_1d_to_1d(telluric['linear']['input']['wave'],
                                   telluric['linear']['input']['step'],
                                   telluric['linear']['input']['coef'],
                                   processed['linear']['wave_sel'],
                                   processed['linear']['step_sel'],
                                   preserve_flux=False)

                """ Saving the outcome to dictionary """
                linear_rebinned_flux *= rescaling_factor
                linear_rebinned_flux += processed['telluric']['spectrum_noairmass'][tel_selection]

                e2ds_corrected = []

                for obs in lists['telluric']:

                    e2ds_corrected.extend(processed['linear'][obs] /
                                            np.power(linear_rebinned_flux, input_data[obs]['AIRMASS']))

                computed_std[n_rescaling_factor] = np.nanstd(e2ds_corrected)
                label_dict = '{0}'.format(rescaling_factor)
                processed['linear']['array_data'][label_dict] = e2ds_corrected
                processed['linear']['array_move'][label_dict] = rescaling_factor


                #print(n_rescaling_factor, rescaling_factor, computed_std[n_rescaling_factor])
                #plt.scatter(processed['linear']['array_wave'], e2ds_corrected)
                #plt.plot(processed['linear']['wave_sel'], linear_rebinned_flux)
                #plt.show()


            #plt.scatter(rescaling_array, computed_std)
            #plt.show()

            """ Selection of the PWV value with the lowest scatter """
            ind_factor = np.argmin(computed_std)
            ind_range = 3
            if ind_factor < ind_range:
                sel_factor = rescaling_array[0:ind_factor+ind_range]
                sel_stdev = computed_std[0:ind_factor+ind_range]
            elif ind_factor > len(rescaling_array) - ind_range:
                sel_factor = rescaling_array[ind_factor-ind_range:]
                sel_stdev = computed_std[ind_factor-ind_range:]
            else:
                sel_factor = rescaling_array[ind_factor-ind_range:ind_factor+ind_range]
                sel_stdev = computed_std[ind_factor-ind_range:ind_factor+ind_range]

            coeff = np.polyfit(sel_factor, sel_stdev, 2)

            PWV_value = - coeff[1] / (2*coeff[0])

            print('   PWV value      : {0:7f}'.format(PWV_value))
            print()

            processed['linear']['PWV_value'] = PWV_value
            processed['linear']['PWV_closest'] = sel_factor
            processed['linear']['computed_std'] = computed_std
            processed['linear']['polyfit'] = {
                'package': 'numpy', # in case we forget what we used...
                'order': 2,
                'coeff': coeff,
                'sel_factor': sel_factor,
                'sel_stdev': sel_stdev
             }

            """ 
            The linear coefficient for the PWV-dependent part are rebinned onto the 2D  scale of the observations. 
            Then the teluric spectrum is computed, using also the baseline from the previous step 
            """

            processed['linear']['rebinned'] = {}
            processed['linear']['rebinned']['coef'] = \
                rebin_1d_to_2d(telluric['linear']['input']['wave'],
                               telluric['linear']['input']['step'],
                               telluric['linear']['input']['coef'],
                               telluric['rebinned']['wave'],
                               telluric['rebinned']['step'],
                               preserve_flux=False)

            processed['linear']['rebinned']['cerr'] = \
                rebin_1d_to_2d(telluric['linear']['input']['wave'],
                               telluric['linear']['input']['step'],
                               telluric['linear']['input']['cerr'],
                               telluric['rebinned']['wave'],
                               telluric['rebinned']['step'],
                               preserve_flux=False,
                               is_error=True)

            sel_out_of_range = ~((telluric['rebinned']['wave'] > telluric['linear']['input']['range'][0]+1.) \
                                & (telluric['rebinned']['wave'] < telluric['linear']['input']['range'][1]-1.))
            processed['linear']['rebinned']['coef'][sel_out_of_range] = 0.
            processed['linear']['rebinned']['cerr'][sel_out_of_range] = 0.1

            processed['telluric']['spectrum_noairmass'] += processed['linear']['rebinned']['coef'] * PWV_value
            processed['telluric']['spectrum_noairmass_err'] = np.sqrt(
                processed['telluric']['spectrum_noairmass_err']**2 +
                (processed['linear']['rebinned']['cerr'] * PWV_value)**2)


        except MissingFileException:
            print(' *** Missing telluric_linear_coeff keyword in configuration file ***')
            print()

        except OutOfRangeException:
            print(' *** Wavelength range for the calculation of the PWV-dependent telluric spectrum is outside the boundaries ***')
            print()

        for obs in lists['observations']:
            """ Correction of telluric lines for the average airmass value, following Wyttenbach et al. 2015 """
            processed[obs]['e2ds_corrected'] = processed[obs]['e2ds_rescaled'] / \
                                                    np.power(processed['telluric']['spectrum_noairmass'],
                                                             input_data[obs]['AIRMASS'])
            processed[obs]['e2ds_corrected_err'] = processed[obs]['e2ds_rescaled_err'] / \
                                                    np.power(processed['telluric']['spectrum_noairmass'],
                                                             input_data[obs]['AIRMASS'])

        for obs in lists['observations']:
            # Correction of telluric lines

            telluric[obs] = {}

            telluric[obs]['spectrum_noairmass'] = processed['telluric']['spectrum_noairmass']

            telluric[obs]['airmass'] = input_data[obs]['AIRMASS']
            telluric[obs]['airmass_ref'] = processed['airmass_ref']

            """ Set anomalosly low point to one (e.g. when the template is not computed)"""
            telluric[obs]['null'] = telluric[obs]['spectrum_noairmass'] < 0.001
            telluric[obs]['spectrum_noairmass'][telluric[obs]['null']] = 1.0

            telluric[obs]['spectrum'] = np.power(processed['telluric']['spectrum_noairmass'],
                                       input_data[obs]['AIRMASS'] - processed['airmass_ref'])

            telluric[obs]['spline_noairmass'] = telluric[obs]['spectrum_noairmass'].copy()

            """ No need to compute the spline approximation since we are already dealing with a very high SNR template"""
            telluric[obs]['spline'] = np.power(telluric[obs]['spline_noairmass'],
                                       input_data[obs]['AIRMASS'] - processed['airmass_ref'])

            """ copy the keyword for future use"""
            telluric[obs]['airmass'] = input_data[obs]['AIRMASS']

            telluric[obs]['telluric_corrected'] = processed[obs]['e2ds_corrected']
            telluric[obs]['telluric_corrected_err'] = processed[obs]['e2ds_corrected_err']

            save_to_cpickle('telluric', telluric, config_in['output'], night)
            save_to_cpickle('telluric_processed', processed, config_in['output'], night)

        print()
        print("Night ", night, " completed")


def plot_telluric_template_alternative(config_in, night_input=''):
    import matplotlib.pyplot as plt

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    system_dict = from_config_get_system(config_in)

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        #plt.scatter(rescaling_array, computed_std, c='C0', zorder=1)
        #plt.scatter(sel_factor, sel_stdev, c='C1', zorder=2)
        #plt.plot(rescaling_array, np.polyval(coeff, rescaling_array))
        #plt.plot(rescaling_array, 2*rescaling_array*coeff[0] + coeff[1] )
        #plt.plot()

        print("plot_telluric_template                     Night: ", night)

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        """ Retrieving the analysis"""
        try:
            processed = load_from_cpickle('telluric_processed', config_in['output'], night)
            telluric = load_from_cpickle('telluric', config_in['output'], night)
        except:
            print()
            print("No telluric correction, no plots")
            continue

        colors, cmap, line_colors = make_color_array(lists, observational_pams)

        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        lift_spectrum = 0.25

        for i, obs in enumerate(lists['observations']):
            color_array = cmap(i / len(lists['observations']))

            _, e2ds_rescaled , _ = \
                perform_rescaling(processed[obs]['wave'],
                                  processed[obs]['e2ds'],
                                  processed[obs]['e2ds_err'],
                                  observational_pams['wavelength_rescaling'])

            e2ds_rescaled_corrected_spectrum = e2ds_rescaled / telluric[obs]['spectrum']
            e2ds_rescaled_corrected_spline = e2ds_rescaled / telluric[obs]['spline']

            for order in range(0, processed[obs]['n_orders']):

                if order == 0 and i==0:
                    ax1.plot(processed[obs]['wave'][order, :],
                                e2ds_rescaled[order, :],
                                c=color_array, lw=1, alpha=0.5, label='uncorrected')
                    ax1.scatter(processed[obs]['wave'][order, :],
                                e2ds_rescaled_corrected_spectrum[order, :],
                                s=1, c=np.atleast_2d(color_array), label='corrected')
                else:
                    ax1.plot(processed[obs]['wave'][order, :],
                                e2ds_rescaled[order, :],
                                c=color_array, lw=1, alpha=0.5)
                    ax1.scatter(processed[obs]['wave'][order, :],
                                e2ds_rescaled_corrected_spectrum[order, :],
                                s=1, c=np.atleast_2d(color_array))

                #ax1.plot(processed[obs]['wave'][order, :],
                #            e2ds_rescaled[order, :]+lift_spectrum,
                #            c=color_array, lw=1, alpha=0.5)
                #ax1.scatter(processed[obs]['wave'][order, :],
                #            e2ds_rescaled_corrected_spline[order, :]+lift_spectrum,
                #            s=1, c=np.atleast_2d(color_array))

                ax2.plot(processed[obs]['wave'][order, :],
                         telluric[obs]['spectrum'][order, :],
                         c=color_array)
                ax2.axhline(1.00, c='k')

                #ax2.plot(processed[obs]['wave'][order, :],
                #         telluric[obs]['spline'][order, :]+lift_spectrum,
                #         c=color_array)
                #ax2.axhline(1.00+lift_spectrum, c='k')

        #ax2.plot(input_data['coadd']['wave'],telluric['stellarRF']['spline_eval']+0.1,c='k')
        #ax2.scatter(input_data['coadd']['wave'],telluric['stellarRF']['spectrum']+0.1,c='r', s=2)

        ax1.legend(loc=3)
        ax1.set_title('Night: ' + night)

        ax2.set_xlabel('$\lambda$ [$\AA$]')

        try:
            instrument = night_dict[night]['instrument']
            comparison_file = config_in['instruments'][instrument]['telluric_comparison']
            comparison_data = np.genfromtxt(comparison_file, skip_header=1)
            if comparison_data[0,0]<1000.0:
                nm2Ang = 10.
            else:
                nm2Ang = 1.
            ax1.plot(comparison_data[:, 0]*nm2Ang, comparison_data[:, 1], c='C0', zorder=1000)
            ax2.plot(comparison_data[:, 0]*nm2Ang, comparison_data[:, 1], c='C0', zorder=1000)
        except:
            pass

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=colors[0], vmax=colors[-1]))
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()


from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.shortcuts import *

__all__ = ["compute_telluric_template",
           "plot_telluric_template",
           "compute_telluric_template_reference",
           "plot_telluric_template_reference"]

def compute_telluric_template(config_in):

    compute_telluric_template_routine(config_in,
                                n_iterations=1,
                                use_berv=True,
                                use_reference_airmass=False,
                                use_template=True,
                                subroutine_name='telluric_template')


def compute_telluric_template_reference(config_in):

    compute_telluric_template_routine(config_in,
                                n_iterations=1,
                                use_berv=True,
                                use_reference_airmass=True,
                                use_template=True,
                                subroutine_name='telluric_template')


def plot_telluric_template_reference(config_in, night_input=''):

    plot_telluric_template(config_in, night_input=night_input)


def compute_telluric_template_routine(config_in, **kwargs):
    """
    Lazy workaround
    :param config_in:
    :param kwargs:
    :return:
    """

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)

    for night in night_dict:

        instrument_name = night_dict[night]['instrument']
        template_dict = instrument_dict[instrument_name]['telluric_template']
        print()
        print("compute_telluric_template                  Night: ", night)
        print()

        try:
            telluric = load_from_cpickle('telluric', config_in['output'], night)
            continue
        except:
            print("  No telluric correction file found, computing now ")
            print()

        print('    instrument :', instrument_name)
        print('    template   :', template_dict['file'])
        print('    fit_range  :', template_dict['fit_range'])
        print()

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        """ Retrieving the observations"""
        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'], use_telluric=False)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        processed = {
            'subroutine': kwargs['subroutine_name'],
            'n_orders': 0,
            'n_pixels': 0
        }

        telluric = {
            'subroutine': kwargs['subroutine_name'],
            'reference_frame': 'observer'
        }

        """ Retrieving the template data"""
        telluric_template_data = np.genfromtxt(template_dict['file'])

        obs_reference = lists['observations'][0]
        telluric['template'] = {
            'input': {
                'range': [np.amin(telluric_template_data[:, 0]), np.amax(telluric_template_data[-1, 0])],
                'wave': telluric_template_data[:, 0],
                'flux': telluric_template_data[:, 1],
                'ferr': telluric_template_data[:, 2],
                'step': telluric_template_data[:, 3]
            },
            'rebinned':{
                'wave': input_data[obs_reference]['wave'],
                'step': input_data[obs_reference]['step']
            }
        }

        """ Reference airmass for iterative correction of airmass"""
        if kwargs['use_reference_airmass']:
            airmass_temp = np.zeros(lists['n_transit_in'])
            for n_obs, obs in enumerate(lists['transit_in']):
                # This is to ensure that airmass, berv and rvc are associated to the correct spectra
                airmass_temp[n_obs] = input_data[obs]['AIRMASS']
            processed['airmass_ref'] = np.average(airmass_temp)
        else:
            processed['airmass_ref'] = 0.000
        processed['telluric'] = {}

        airmass = np.zeros(lists['n_observations'], dtype=np.double)
        berv = np.zeros(lists['n_observations'], dtype=np.double)
        rvc = np.zeros(lists['n_observations'], dtype=np.double)

        # There must be a more elegant way to do this, but I'm, not aware of it
        for n_obs, obs in enumerate(lists['observations']):

            tel_selection = (input_data[obs]['wave'] > template_dict['fit_range'][0]) & \
                            (input_data[obs]['wave'] < template_dict['fit_range'][1])

            processed[obs] = {
                'n_orders': input_data[obs]['n_orders'],
                'n_pixels': input_data[obs]['n_pixels']
            }

            """ for plotting purpose only"""
            processed[obs]['wave'] = input_data[obs]['wave']
            processed[obs]['e2ds'] = input_data[obs]['e2ds']
            processed[obs]['e2ds_err'] = input_data[obs]['e2ds_err']


            processed[obs]['e2ds_rescaling'], processed[obs]['e2ds_rescaled'], processed[obs]['e2ds_rescaled_err'] = \
                perform_rescaling(input_data[obs]['wave'],
                                  input_data[obs]['e2ds'],
                                  input_data[obs]['e2ds_err'],
                                  observational_pams['wavelength_rescaling'])

            processed[obs]['e2ds_sel'] = processed[obs]['e2ds_rescaled'][tel_selection]

            if processed['n_orders'] == 0:
                processed['wave_sel'] = input_data[obs]['wave'][tel_selection]
                processed['step_sel'] = input_data[obs]['step'][tel_selection]
                processed['n_orders'] = input_data[obs]['orders']
                processed['n_pixels'] = input_data[obs]['wave_size']

            # This is to ensure that airmass, berv and rvc are associated to the correct spectra
            processed['telluric'][obs] = {'n_obs': n_obs}
            airmass[n_obs] = input_data[obs]['AIRMASS']
            berv[n_obs] = input_data[obs]['BERV']
            rvc[n_obs] = input_data[obs]['RVC']

        processed['template_fit'] = {}

        rescaling_array = np.arange(0.05, 2.0, 0.05)
        computed_std = np.zeros(len(rescaling_array))

        processed['template_fit']['array_data'] = {}
        processed['template_fit']['array_move'] = {}
        processed['template_fit']['array_wave'] = []

        """ saving the wavelength array for plotting purpose"""
        for obs in lists['telluric']:
            processed['template_fit']['array_wave'].extend(processed['wave_sel'])

        for n_rescaling_factor, rescaling_factor in enumerate(rescaling_array):

            """ Template spectrum is rebinned onto the observations wavelength scale, using only the wavelength range
            selected for the computation of the rescaling factor
            """
            template_rebinned_flux = \
                rebin_1d_to_1d(telluric['template']['input']['wave'],
                               telluric['template']['input']['step'],
                               telluric['template']['input']['flux'],
                               processed['wave_sel'],
                               processed['step_sel'],
                               preserve_flux=False)

            """ Saving the outcome to dictionary """
            template_rebinned_flux -= 1.00
            template_rebinned_flux *= rescaling_factor
            template_rebinned_flux += 1.00

            e2ds_corrected = []

            ## DEBUG
            # wave_corrected = []
            # e2ds_original = []

            for obs in lists['telluric']:

                e2ds_corrected.extend(processed[obs]['e2ds_sel'] /
                                        np.power(template_rebinned_flux, input_data[obs]['AIRMASS']))

                # wave_corrected.extend(processed['wave_sel'])
                # e2ds_original.extend(processed[obs]['e2ds_sel'])

            computed_std[n_rescaling_factor] = np.std(e2ds_corrected)

            label_dict = '{0}'.format(rescaling_factor)
            # print(label_dict, computed_std[n_rescaling_factor])

            # processed['template_fit']['array_data'][label_dict] = e2ds_corrected
            # processed['template_fit']['array_move'][label_dict] = rescaling_factor
            # plt.scatter(wave_corrected, e2ds_original, s=2, alpha=0.5)
            # plt.scatter(wave_corrected, e2ds_corrected, s=2, alpha=0.5)
            # plt.plot(processed['wave_sel'],template_rebinned_flux)
            # plt.show()

        """ selection of the rescaling factor with the lowest scatter """
        ind_factor = np.argmin(computed_std)
        ind_range = 3
        if ind_factor < ind_range:
            sel_factor = rescaling_array[0:ind_factor+ind_range]
            sel_stdev = computed_std[0:ind_factor+ind_range]
        elif ind_factor > len(rescaling_array) - ind_range:
            sel_factor = rescaling_array[ind_factor-ind_range:]
            sel_stdev = computed_std[ind_factor-ind_range:]
        else:
            sel_factor = rescaling_array[ind_factor-ind_range:ind_factor+ind_range]
            sel_stdev = computed_std[ind_factor-ind_range:ind_factor+ind_range]

        coeff = np.polyfit(sel_factor, sel_stdev, 2)

        telluric_factor = - coeff[1] / (2*coeff[0])

        print('   telluric factor: {0:7f}'.format(telluric_factor))
        print()

        processed['template_fit']['telluric_factor'] = telluric_factor
        processed['template_fit']['rescaling_factor'] = rescaling_factor
        processed['template_fit']['computed_std'] = computed_std
        processed['template_fit']['polyfit'] = {
            'package': 'numpy', # in case we forget what we used...
            'order': 2,
            'coeff': coeff,
            'sel_factor': sel_factor,
            'sel_stdev': sel_stdev
         }

        """ After being rescaled for the proper factor, the template telluric spectrum is rebinned onto the 2D
        scale of the observations """

        telluric['template']['rebinned']['flux'] = \
            rebin_1d_to_2d(telluric['template']['input']['wave'],
                           telluric['template']['input']['step'],
                           telluric['template']['input']['flux'],
                           telluric['template']['rebinned']['wave'],
                           telluric['template']['rebinned']['step'],
                           preserve_flux=False)

        telluric['template']['rebinned']['ferr'] = \
            rebin_1d_to_2d(telluric['template']['input']['wave'],
                           telluric['template']['input']['step'],
                           telluric['template']['input']['ferr'],
                           telluric['template']['rebinned']['wave'],
                           telluric['template']['rebinned']['step'],
                           preserve_flux=False,
                           is_error=True)

        sel_out_of_range = ~((telluric['template']['rebinned']['wave'] > telluric['template']['input']['range'][0]+1.) \
                            & (telluric['template']['rebinned']['wave'] < telluric['template']['input']['range'][1]-1.))
        telluric['template']['rebinned']['flux'][sel_out_of_range] = 1.
        telluric['template']['rebinned']['ferr'][sel_out_of_range] = 0.1


        processed['telluric']['spectrum_noairmass'] = \
            (telluric['template']['rebinned']['flux'] - 1.) * telluric_factor + 1.0

        telluric['airmass_ref'] = processed['airmass_ref']

        for obs in lists['observations']:
            """ Correction of telluric lines for the average airmass value, following Wyttenbach et al. 2015 """
            processed[obs]['e2ds_corrected'] = processed[obs]['e2ds_rescaled'] / \
                                                    np.power(processed['telluric']['spectrum_noairmass'],
                                                             input_data[obs]['AIRMASS'] - processed['airmass_ref'])
            processed[obs]['e2ds_corrected_err'] = processed[obs]['e2ds_rescaled_err'] / \
                                                    np.power(processed['telluric']['spectrum_noairmass'],
                                                             input_data[obs]['AIRMASS'] - processed['airmass_ref'])

        for obs in lists['observations']:
            # Correction of telluric lines

            telluric[obs] = {}

            telluric[obs]['spectrum_noairmass'] = processed['telluric']['spectrum_noairmass']

            telluric[obs]['airmass'] = input_data[obs]['AIRMASS']
            telluric[obs]['airmass_ref'] = processed['airmass_ref']

            """ Set anomalosly low point to one (e.g. when the template is not computed)"""
            telluric[obs]['null'] = telluric[obs]['spectrum_noairmass'] < 0.001
            telluric[obs]['spectrum_noairmass'][telluric[obs]['null']] = 1.0

            telluric[obs]['spectrum'] = np.power(processed['telluric']['spectrum_noairmass'],
                                       input_data[obs]['AIRMASS'] - processed['airmass_ref'])

            telluric[obs]['spline_noairmass'] = telluric[obs]['spectrum_noairmass'].copy()

            """ No need to compute the spline approximation since we are already dealing with a very high SNR template"""
            telluric[obs]['spline'] = np.power(telluric[obs]['spline_noairmass'],
                                       input_data[obs]['AIRMASS'] - processed['airmass_ref'])

            """ copy the keyword for future use"""
            telluric[obs]['airmass'] = input_data[obs]['AIRMASS']

            telluric[obs]['telluric_corrected'] = processed[obs]['e2ds_corrected']
            telluric[obs]['telluric_corrected_err'] = processed[obs]['e2ds_corrected_err']

            save_to_cpickle('telluric', telluric, config_in['output'], night)
            save_to_cpickle('telluric_processed', processed, config_in['output'], night)

        print()
        print("Night ", night, " completed")


def plot_telluric_template(config_in, night_input=''):
    import matplotlib.pyplot as plt

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    system_dict = from_config_get_system(config_in)

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        #plt.scatter(rescaling_array, computed_std, c='C0', zorder=1)
        #plt.scatter(sel_factor, sel_stdev, c='C1', zorder=2)
        #plt.plot(rescaling_array, np.polyval(coeff, rescaling_array))
        #plt.plot(rescaling_array, 2*rescaling_array*coeff[0] + coeff[1] )
        #plt.plot()

        print("plot_telluric_template                     Night: ", night)

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        """ Retrieving the analysis"""
        try:
            processed = load_from_cpickle('telluric_processed', config_in['output'], night)
            telluric = load_from_cpickle('telluric', config_in['output'], night)
        except:
            print()
            print("No telluric correction, no plots")
            continue

        colors, cmap, line_colors = make_color_array(lists, observational_pams)

        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        lift_spectrum = 0.25

        for i, obs in enumerate(lists['observations']):
            color_array = cmap(i / len(lists['observations']))

            _, e2ds_rescaled , _ = \
                perform_rescaling(processed[obs]['wave'],
                                  processed[obs]['e2ds'],
                                  processed[obs]['e2ds_err'],
                                  observational_pams['wavelength_rescaling'])

            e2ds_rescaled_corrected_spectrum = e2ds_rescaled / telluric[obs]['spectrum']
            e2ds_rescaled_corrected_spline = e2ds_rescaled / telluric[obs]['spline']

            for order in range(0, processed[obs]['n_orders']):

                if order == 0 and i==0:
                    ax1.plot(processed[obs]['wave'][order, :],
                                e2ds_rescaled[order, :],
                                c=color_array, lw=1, alpha=0.5, label='uncorrected')
                    ax1.scatter(processed[obs]['wave'][order, :],
                                e2ds_rescaled_corrected_spectrum[order, :],
                                s=1, c=np.atleast_2d(color_array), label='corrected')
                else:
                    ax1.plot(processed[obs]['wave'][order, :],
                                e2ds_rescaled[order, :],
                                c=color_array, lw=1, alpha=0.5)
                    ax1.scatter(processed[obs]['wave'][order, :],
                                e2ds_rescaled_corrected_spectrum[order, :],
                                s=1, c=np.atleast_2d(color_array))

                #ax1.plot(processed[obs]['wave'][order, :],
                #            e2ds_rescaled[order, :]+lift_spectrum,
                #            c=color_array, lw=1, alpha=0.5)
                #ax1.scatter(processed[obs]['wave'][order, :],
                #            e2ds_rescaled_corrected_spline[order, :]+lift_spectrum,
                #            s=1, c=np.atleast_2d(color_array))

                ax2.plot(processed[obs]['wave'][order, :],
                         telluric[obs]['spectrum'][order, :],
                         c=color_array)
                ax2.axhline(1.00, c='k')

                #ax2.plot(processed[obs]['wave'][order, :],
                #         telluric[obs]['spline'][order, :]+lift_spectrum,
                #         c=color_array)
                #ax2.axhline(1.00+lift_spectrum, c='k')

        #ax2.plot(input_data['coadd']['wave'],telluric['stellarRF']['spline_eval']+0.1,c='k')
        #ax2.scatter(input_data['coadd']['wave'],telluric['stellarRF']['spectrum']+0.1,c='r', s=2)

        ax1.legend(loc=3)
        ax1.set_title('Night: ' + night)

        ax2.set_xlabel('$\lambda$ [$\AA$]')

        try:
            instrument = night_dict[night]['instrument']
            comparison_file = config_in['instruments'][instrument]['telluric_comparison']
            comparison_data = np.genfromtxt(comparison_file, skip_header=1)
            if comparison_data[0,0]<1000.0:
                nm2Ang = 10.
            else:
                nm2Ang = 1.
            ax1.plot(comparison_data[:, 0]*nm2Ang, comparison_data[:, 1], c='C0', zorder=1000)
            ax2.plot(comparison_data[:, 0]*nm2Ang, comparison_data[:, 1], c='C0', zorder=1000)
        except:
            pass

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=colors[0], vmax=colors[-1]))
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()



from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *

from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.shortcuts import *

__all__ = ["compute_interstellar_lines", "plot_interstellar_lines"]

subroutine_name = 'interstellar_lines'


#def plot_identify_stellar_lines(config_in)

def compute_interstellar_lines(config_in):

    night_dict = from_config_get_nights(config_in)

    interstellar_lines = from_config_get_interstellar_lines(config_in)

    if not interstellar_lines:
        return

    for night in night_dict:

        print()
        print("compute_interstellar_lines                 Night: ", night)

        try:
            interstellar = load_from_cpickle('interstellar_lines', config_in['output'], night)
            continue
        except:
            print()
            print("  No interstellar correction file found, computing now ")

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'])
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        processed = {
            'subroutine': subroutine_name
        }

        interstellar = {
            'subroutine': subroutine_name,
        }

        import matplotlib.pyplot as plt

        for obs in lists['observations']:

            processed[obs] = {
                'n_orders': input_data[obs]['n_orders'],
                'n_pixels': input_data[obs]['n_pixels']
            }
            interstellar[obs] = {}

            """ for plotting purpose only"""
            processed[obs]['wave'] = input_data[obs]['wave']

            processed[obs]['flux'] = input_data[obs]['e2ds']/calib_data['blaze']/input_data[obs]['step']
            processed[obs]['flux_err'] = np.sqrt(input_data[obs]['e2ds'])/calib_data['blaze']/input_data[obs]['step']

            if obs in lists['telluric']:

                try:
                    interstellar['flux_total'] += processed[obs]['flux']
                    interstellar['flux_total_err'] += processed[obs]['flux_err']**2
                except:
                    interstellar['wave'] = input_data[obs]['wave']
                    interstellar['flux_total'] = processed[obs]['flux'][:,:]
                    interstellar['flux_total_err'] = processed[obs]['flux_err']**2

        interstellar['flux_total_err'] = np.sqrt(interstellar['flux_total_err'])

        """ Zero or negative values are identified, flagged and substituted with another value """
        interstellar['flux_total'], interstellar['flux_total_err'], interstellar['null'] = \
                replace_values_errors(interstellar['flux_total'], interstellar['flux_total_err'], 0.0001)

        """rescaling"""
        interstellar['flux_rescaling'], interstellar['flux_rescaled'],interstellar['flux_rescaled_err']  = \
                perform_rescaling(interstellar['wave'],
                                  interstellar['flux_total'],
                                  interstellar['flux_total_err'],
                                  observational_pams['wavelength_rescaling'])

        interstellar['correction'] = np.ones(np.shape(interstellar['wave']))

        for line_name, line in interstellar_lines.items():

            interstellar[line_name] = {}

            sel1 = (np.abs(interstellar['wave']-line[0])<line[1])
            sel2 = (~sel1) & (np.abs(interstellar['wave']-line[0])<line[2])
            sel3 = (sel1 | sel2)

            poly_coeff = np.polyfit(interstellar['wave'][sel2], interstellar['flux_rescaled'][sel2], 2)

            normalized = interstellar['flux_rescaled'][sel3]/np.polyval(poly_coeff, interstellar['wave'][sel3])

            interstellar[line_name]['spline_eval'], \
            interstellar[line_name]['spline_coeff'], \
            interstellar[line_name]['spline_knots'] = \
                compute_spline(interstellar['wave'][sel3], normalized, 0.04)

            interstellar['correction'][sel1] = sci_int.splev(interstellar['wave'][sel1], interstellar[line_name]['spline_coeff'])

            interstellar[line_name]['sel1'] = sel1
            interstellar[line_name]['sel2'] = sel2
            interstellar[line_name]['sel3'] = sel3

        save_to_cpickle('interstellar_lines_processed', processed, config_in['output'], night)
        save_to_cpickle('interstellar_lines', interstellar, config_in['output'], night)


def plot_interstellar_lines(config_in, night_input=''):
    import matplotlib.pyplot as plt

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    system_dict = from_config_get_system(config_in)

    interstellar_lines = from_config_get_interstellar_lines(config_in)

    if not interstellar_lines:
        return

    if night_input=='':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        print("plot_interstellar_lines                    Night: ", night)
        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        """ Retrieving the analysis"""
        try:
            processed = load_from_cpickle('interstellar_lines_processed', config_in['output'], night)
            interstellar = load_from_cpickle('interstellar_lines', config_in['output'], night)
        except:
            print()
            print('No interstellar correction, no plots')
            continue

        colors, cmap, line_colors = make_color_array(lists, observational_pams)

        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        for i, obs in enumerate(lists['observations']):

            """rescaling"""
            processed[obs]['flux_rescaling'], processed[obs]['flux_rescaled'], processed[obs]['flux_rescaled_err']  = \
                perform_rescaling(interstellar['wave'],
                                  processed[obs]['flux'],
                                  processed[obs]['flux_err'],
                                  observational_pams['wavelength_rescaling'])
            ax1.scatter(interstellar['wave'], processed[obs]['flux_rescaled'],
                            s=1, c=line_colors[i])

            #ax1.plot(interstellar['wave'], interstellar['correction'], c='black')

            ax2.scatter(interstellar['wave'], processed[obs]['flux_rescaled']/interstellar['correction'],
                        s=1, c=line_colors[i])

        for line_name, line in interstellar_lines.items():

            #ax1.axvline(line[0], c='k')
            ax1.axvline(line[0]-line[1], c='b')
            ax1.axvline(line[0]+line[1], c='b')
            ax1.axvline(line[0]-line[2], c='g')
            ax1.axvline(line[0]+line[2], c='g')

            #ax2.axvline(line[0], c='k')
            ax2.axvline(line[0]-line[1], c='b')
            ax2.axvline(line[0]+line[1], c='b')
            ax2.axvline(line[0]-line[2], c='g')
            ax2.axvline(line[0]+line[2], c='g')

            try:
                wave_min = min(wave_min, line[0])
                wave_max = max(wave_max, line[0])
                range_max = max(range_max, line[2])
            except:
                wave_min = line[0]
                wave_max = line[0]
                range_max = line[2]

        #ax1.legend(loc=3)
        ax1.set_title('Night: ' + night)
        ax1.set_xlim(wave_min-2*range_max, wave_max+2*range_max)

        ax2.set_xlabel('$\lambda$ [$\AA$]')
        ax1.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=colors[0], vmax=colors[-1]))
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()


from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.shortcuts import *
from SLOPpy.subroutines.rebin_subroutines import *
from astropy.convolution import convolve, Box1DKernel

__all__ = ['compute_master_out', 'plot_master_out', 'plot_compare_master_out']
subroutine_name = 'master_out'


def compute_master_out(config_in):

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    system_dict = from_config_get_system(config_in)

    shared_data = load_from_cpickle('shared', config_in['output'])

    master_out_composite = {
        'subroutine': 'master_out',
        'wave': shared_data['coadd']['wave'],
        'step': shared_data['coadd']['step'],
        'size': shared_data['coadd']['size'],
    }
    wmean_wflux = np.zeros(master_out_composite['size'])
    wmean_weight = np.zeros(master_out_composite['size'])

    box_kernel = Box1DKernel(config_in['master-out'].get('boxcar_smoothing', 1))

    for night in night_dict:

        try:
            master_out = load_from_cpickle('master_out', config_in['output'], night)
            wmean_wflux += master_out['rescaled'] / master_out['rescaled_err'] ** 2
            wmean_weight += 1. // master_out['rescaled_err'] ** 2
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Retrieved'))
            continue
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Computing'))
            print()

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'])
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        processed = {
            'subroutine': 'master_out'
        }

        master_out = {
            'subroutine': subroutine_name,
            'wave': shared_data['coadd']['wave'],
            'step': shared_data['coadd']['step'],
            'size': shared_data['coadd']['size'],
            'total_flux': np.zeros(shared_data['coadd']['size'], dtype=np.double),
            'total_flux_err': np.zeros(shared_data['coadd']['size'], dtype=np.double)
        }

        for obs in lists['transit_out']:

            processed[obs] = {}

            processed[obs]['rescaling'], \
            processed[obs]['rescaled'], \
            processed[obs]['rebinned_err'] = perform_rescaling(
                input_data[obs]['wave'], input_data[obs]['e2ds'], input_data[obs]['e2ds_err'],
                observational_pams['wavelength_rescaling'])

            preserve_flux = input_data[obs].get('absolute_flux', True)

            processed[obs]['rebinned'] = \
                rebin_2d_to_1d(input_data[obs]['wave'],
                               input_data[obs]['step'],
                               input_data[obs]['e2ds'],
                               calib_data['blaze'],
                               master_out['wave'],
                               master_out['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2SRF_mod'])

            processed[obs]['rebinned_err'] = \
                rebin_2d_to_1d(input_data[obs]['wave'],
                               input_data[obs]['step'],
                               input_data[obs]['e2ds_err'],
                               calib_data['blaze'],
                               master_out['wave'],
                               master_out['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2SRF_mod'])

            master_out['total_flux'] += processed[obs]['rebinned']
            master_out['total_flux_err'] += processed[obs]['rebinned_err']**2.0

        master_out['total_flux_err'] = np.sqrt(master_out['total_flux_err'])

        master_out['rescaling'], \
        master_out['rescaled'], \
        master_out['rescaled_err'] = perform_rescaling(
            master_out['wave'], master_out['total_flux'],
            master_out['total_flux_err'],
            observational_pams['wavelength_rescaling'])

        master_out['rescaled'], master_out['rescaled_err'], master_out['null'] = \
            replace_values_errors(master_out['rescaled'], master_out['rescaled_err'],
                                  threshold=0.0001, replacement=1.0000)

        master_out['smoothed'] = convolve(master_out['rescaled'].copy(), box_kernel)
        master_out['smoothed_err'] = np.sqrt(convolve((master_out['rescaled_err'])**2, box_kernel))

        sel = (master_out['smoothed']<0.01) | (master_out['smoothed']>1.5)
        master_out['smoothed'][sel] = 1.0
        master_out['smoothed_err'][sel] = 1.0

        selection = (master_out['wave']>0)
        spline_iter = 5
        for n_iter in range(0, spline_iter):

            residuals = master_out['rescaled'] / master_out['smoothed']
            wave = master_out_composite['wave']

            """ picking the number of knots """
            nknots = ((np.amax(wave) - np.amin(wave)) / config_in['master-out'].get('spline_step', 0.10))
            """ picking the indices of the knots"""
            idx_knots = (np.arange(1, len(wave[selection]) - 1, (len(wave[selection]) - 2.) / nknots)).astype('int')
            """ passing from indices to knots values """
            knots = wave[selection][idx_knots]

            coeff = sci_int.splrep(wave[selection], residuals[selection], task=-1, k=2, t=knots)
            spline = sci_int.splev(wave, coeff)

            dif = residuals - spline
            std = np.std(dif)
            selection = np.where(np.abs(dif) < 4 * std)  # & (refraction[obs]['flag'])

        master_out['spline'] = spline
        master_out['smoothed'] *= spline
        master_out['smoothed_err'] *= spline

        master_out['SRF'] = {}

        master_out['SRF']['rescaled']= \
            rebin_1d_to_1d(master_out['wave'],
                           master_out['step'],
                           master_out['rescaled'],
                           master_out['wave'], master_out['step'],
                           rv_shift=observational_pams['rv_shift_ORF2SRF_res'],
                           preserve_flux=False)

        master_out['SRF']['rescaled_err']= \
            rebin_1d_to_1d(master_out['wave'],
                           master_out['step'],
                           master_out['rescaled_err'],
                           master_out['wave'], master_out['step'],
                           rv_shift=observational_pams['rv_shift_ORF2SRF_res'],
                           preserve_flux=False,
                           is_error=True)

        wmean_wflux += master_out['SRF']['rescaled']/master_out['SRF']['rescaled_err']**2
        wmean_weight += 1.//master_out['SRF']['rescaled_err']**2



        """
        rv_shift = observational_pams['BERV_avg'] - observational_pams['RV_star']['intercept']
        # bringing the master-out to the aboslute reference system 
        wave_shifted, _ = shift_wavelength(master_out['wave'], master_out['step'], rv_shift)

        # master-out is printed to .dat for compatibility with other programs
        master_data_out = get_filename('master_out', config_in['output'], night, extension=".dat")
        file_out = open(master_data_out, 'w')
        for w, f, e in  zip(wave_shifted, master_out['rescaled'], master_out['rescaled_err']):
            file_out.write('{0:10.4f} {1:f} {2:f}\n'.format(w,f,e))
        file_out.close()
        print()
        print("NON OPTIMAL MASTER-OUT DAT FILE!!!!")
        print()
        """

        save_to_cpickle('master_out_processed', processed, config_in['output'], night)
        save_to_cpickle('master_out', master_out, config_in['output'], night)

    master_out_composite['SRF'] = {}
    master_out_composite['SRF']['rescaled'] = wmean_wflux/wmean_weight
    master_out_composite['SRF']['rescaled_err'] = np.sqrt(1./wmean_weight)
    master_out_composite['SRF']['smoothed'] = convolve(master_out_composite['SRF']['rescaled'].copy(), box_kernel)
    master_out_composite['SRF']['smoothed_err'] = \
        np.sqrt(convolve((master_out_composite['SRF']['rescaled_err']) ** 2, box_kernel))

    print()
    for night in night_dict:
        try:
            master_out_composite = load_from_cpickle('master_out_composite', config_in['output'], night)
            print("{0:45s} Night:{1:15s}   {2:s}".format('master_out_composite', night, 'Retrieved'))
            continue
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format('master_out_composite', night, 'Computing'))
            print()

        master_out = load_from_cpickle('master_out', config_in['output'], night)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        master_out_composite['rescaled']= \
            rebin_1d_to_1d(master_out_composite['wave'],
                           master_out_composite['step'],
                           master_out_composite['SRF']['rescaled'],
                           master_out_composite['wave'],
                           master_out_composite['step'],
                           rv_shift=-observational_pams['rv_shift_ORF2SRF_res'],
                           preserve_flux=False)

        master_out_composite['rescaled_err']= \
            rebin_1d_to_1d(master_out_composite['wave'],
                           master_out_composite['step'],
                           master_out_composite['SRF']['rescaled_err'],
                           master_out_composite['wave'],
                           master_out_composite['step'],
                           rv_shift=-observational_pams['rv_shift_ORF2SRF_res'],
                           preserve_flux=False,
                           is_error=True)

        master_out_composite['smoothed'] = \
            rebin_1d_to_1d(master_out_composite['wave'],
                           master_out_composite['step'],
                           master_out_composite['SRF']['smoothed'],
                           master_out_composite['wave'],
                           master_out_composite['step'],
                           rv_shift=-observational_pams['rv_shift_ORF2SRF_res'],
                           preserve_flux=False)

        master_out_composite['smoothed_err']= \
            rebin_1d_to_1d(master_out_composite['wave'],
                           master_out_composite['step'],
                           master_out_composite['SRF']['smoothed_err'],
                           master_out_composite['wave'],
                           master_out_composite['step'],
                           rv_shift=-observational_pams['rv_shift_ORF2SRF_res'],
                           preserve_flux=False,
                           is_error=True)

        #master_out_composite['smoothed'] = convolve(master_out_composite['rescaled'].copy(), box_kernel)
        #master_out_composite['smoothed_err'] = \
        #    np.sqrt(convolve((master_out_composite['rescaled_err']) ** 2, box_kernel))

        sel = (master_out_composite['smoothed']<0.01) | (master_out_composite['smoothed']>1.5)
        master_out_composite['smoothed'][sel] = 1.0
        master_out_composite['smoothed_err'][sel] = 1.0

        selection = (master_out_composite['wave']>0)
        spline_iter = 5
        for n_iter in range(0, spline_iter):

            residuals = master_out['rescaled'] / master_out_composite['smoothed']
            wave = master_out_composite['wave']

            """ picking the number of knots """
            nknots = ((np.amax(wave) - np.amin(wave)) / config_in['master-out'].get('spline_step', 0.10))
            """ picking the indices of the knots"""
            idx_knots = (np.arange(1, len(wave[selection]) - 1, (len(wave[selection]) - 2.) / nknots)).astype('int')
            """ passing from indices to knots values """
            knots = wave[selection][idx_knots]

            coeff = sci_int.splrep(wave[selection], residuals[selection], task=-1, k=2, t=knots)
            spline = sci_int.splev(wave, coeff)

            dif = residuals - spline
            std = np.std(dif)
            selection = np.where(np.abs(dif) < 4 * std)  # & (refraction[obs]['flag'])

        master_out_composite['spline'] = spline
        master_out_composite['smoothed'] *= spline
        master_out_composite['smoothed_err'] *= spline

        save_to_cpickle('master_out_composite', master_out_composite, config_in['output'], night)


def plot_master_out(config_in, night_input=''):
    import matplotlib.pyplot as plt

    night_dict = from_config_get_nights(config_in)

    if night_input=='':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        print("plot_master_out                            Night: ", night)

        """ Retrieving the analysis"""
        try:
            master_out = load_from_cpickle('master_out', config_in['output'], night)
            master_out_composite = load_from_cpickle('master_out_composite', config_in['output'], night)
        except:
            print()
            print("No master_out , no plots")
            continue

        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        plt.figure(figsize=(12, 6))
        plt.title('Master out - night ' + night)

        lists = load_from_cpickle('lists', config_in['output'], night)
        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'])
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)


        obs = lists['transit_out'][0]
        plt.scatter(input_data[obs]['wave'],
                 calib_data['blaze'],
                 color='C3', zorder=3., label='blaze', alpha=0.25)

        plt.errorbar(master_out['wave'],
                     master_out['rescaled'],
                     yerr=master_out['rescaled_err'],
                     fmt='.', c='C0', label='master-out ' + night)
        plt.plot(master_out['wave'],
                 master_out['smoothed'],
                 color='C1', zorder=3., label='smoothed master-out ' + night)
        plt.scatter(master_out_composite['wave'],
                    master_out_composite['rescaled'],
                    s=2, c='C3')
        plt.plot(master_out_composite['wave'],
                 master_out_composite['smoothed'],
                 c='C3', label='composite master-out')
        plt.scatter(master_out['wave'],
                    master_out['rescaled']/master_out['smoothed']*master_out['spline']+0.05,
                    s=2, c='C4', label='rescaled/smoothed')
        plt.scatter(master_out['wave'],
                    master_out['rescaled']/master_out_composite['smoothed']*master_out_composite['spline']+0.1,
                    s=2, c='C5', label='rescaled/ comp smoothed')
        plt.plot(master_out['wave'],
                 master_out['spline']+0.05,
                 c='C7', label='spline fit of the residuals')
        plt.plot(master_out['wave'],
                 master_out_composite['spline']+0.1,
                 c='C7')
        plt.ylim(0, 1.25)
        plt.xlabel('$\lambda$ [$\AA$]')
        plt.ylabel('Rescaled flux')
        plt.legend()
        plt.show()


def plot_compare_master_out(config_in):

    plt.figure(figsize=(12, 6))
    plt.title('Master out - comparison between nights ')
    night_dict = from_config_get_nights(config_in)

    for i, night in enumerate(night_dict):

        """ Retrieving the analysis"""
        try:
            master_out = load_from_cpickle('master_out', config_in['output'], night)
            master_out_composite = load_from_cpickle('master_out_composite', config_in['output'], night)
        except:
            continue

        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        plt.errorbar(master_out['wave'],
                     master_out['SRF']['rescaled'],
                     yerr=master_out['SRF']['rescaled_err'],
                     fmt='.', c='C'+repr(i), label='master-out ' + night, alpha=0.5)

        if i == 0:
            plt.plot(master_out_composite['wave'],
                     master_out_composite['SRF']['rescaled'],
                     color='k', zorder=10, label='composite master-out')

    plt.ylim(0, 1.25)
    plt.xlabel('$\lambda$ [$\AA$]')
    plt.ylabel('Rescaled flux')
    plt.legend()
    plt.show()


from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.shortcuts import *

__all__ = ["compute_telluric_molecfit_v1_preparation",
           "plot_telluric_molecfit_v1_preparation"]

def compute_telluric_molecfit_v1_preparation(config_in):
    """
    Lazy workaround
    :param config_in:
    :param kwargs:
    :return:
    """

    night_dict = from_config_get_nights(config_in)

    molecfit_dict = from_config_get_molecfit(config_in)

    for night in night_dict:

        try:
            tellprep = load_from_cpickle('telluric_molecfit_preparation', config_in['output'], night)
            continue
        except:
            print()
            print("compute_telluric_molecfit_preparation             Night: ", night)
            print()

        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        tellprep = {
            'work_dir':  config_in['output'] + '_molecfit_' + night,
            'include': {}
        }



        """
        We store all the molecfit files in a subdirectory
        We save the path of the main directory to a temporary file
        """

        os.system('mkdir -p ' + tellprep['work_dir'])
        os.system('mkdir -p ' + tellprep['work_dir'] + '/output/')

        """
        Creation of the include files
        """


        """
        includes_spans_ORF: wavelength ranges _with_ telluric lines, in the ORF
        includes_spans_SRF: wavelength ranges _without_ stellar lines and broadly
            overlapping with telluric ranges, in the SRF
        the two lists must have the same number of columns, with precise correspondence
        """
        tellprep['include']['spans_telluric'] = np.genfromtxt(molecfit_dict['include_telluric'])
        tellprep['include']['spans_stellar_SRF'] = np.genfromtxt(molecfit_dict['include_stellar'])

        #print()
        #print(tellprep['include']['spans_telluric'])
        #print()
        #print(tellprep['include']['spans_stellar_SRF'])

        """ shift the stellar wavelength ranges into ORF """
        tellprep['include']['rv_shift_SRF2ORF'] = -observational_pams['BERV_avg'] + observational_pams['RV_star'][
            'RV_systemic']

        #print()
        #print(tellprep['include']['rv_shift_SRF2ORF'])
        #print(observational_pams['BERV_avg'])
        #print(observational_pams['RV_star']['RV_systemic'])
        #print()
        #print()

        tellprep['include']['spans_stellar'] = tellprep['include']['spans_stellar_SRF']\
                                               * (tellprep['include']['rv_shift_SRF2ORF']
                                                  / (299792458. / 1000.000) + 1.00000)

        #print()
        #print(tellprep['include']['spans_stellar'])

        """ Selecting the overlapping regions between the two lists: we want telluric regions that are not contaminated
            by stellar lines,
        """
        sel_lower = (tellprep['include']['spans_stellar'][:, 0] > tellprep['include']['spans_telluric'][:, 0])
        sel_upper = (tellprep['include']['spans_stellar'][:, 1] < tellprep['include']['spans_telluric'][:, 1])

        """ Final list in the ORF is built"""
        tellprep['include']['selected'] = tellprep['include']['spans_telluric'].copy()
        tellprep['include']['selected'][sel_lower, 0] = tellprep['include']['spans_stellar'][sel_lower, 0]
        tellprep['include']['selected'][sel_upper, 1] = tellprep['include']['spans_stellar'][sel_upper, 1]

        #print()
        #print(tellprep['include']['selected'])

        """ Molecfit line list must be given in vacuum wavelength, even if the stellar spectra is in air wavelength
            conversion from air to vacuum for include file preparation
            where s = 10000 / lambda air and the conversion is: lambda_vac = lambda_air * n.
            http://www.astro.uu.se/valdwiki/Air-to-vacuum%20conversion
        """
        s2 = (10000. / tellprep['include']['selected']) ** 2
        n = 1 + 0.00008336624212083 + 0.02408926869968 / (130.1065924522 - s2) + 0.0001599740894897 / (
                38.92568793293 - s2)
        tellprep['include']['vacuum'] = tellprep['include']['selected'] * n / 10000.

        fileout = open('./' + tellprep['work_dir'] + '/include_' + night + '.dat', 'w')
        for i_s, i_e in zip(tellprep['include']['vacuum'][:, 0], tellprep['include']['vacuum'][:, 1]):
            fileout.write('{0:12.8f} {1:12.8f}\n'.format(i_s, i_e))
        fileout.close()


        #quit()
        save_to_cpickle('telluric_molecfit_preparation', tellprep, config_in['output'], night)



def plot_telluric_molecfit_v1_preparation(config_in, night_input=''):
    import matplotlib.pyplot as plt

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    system_dict = from_config_get_system(config_in)

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:
        print("plot_telluric_template                     Night: ", night)



from __future__ import print_function, division
from SLOPpy.instruments.get_data import *
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.shortcuts import *

__all__ = ["prepare_datasets", "plot_dataset"]


def _check_wavelength_rescaling(wave_rescaling, wave_observations):
    if wave_rescaling[0] < wave_observations[0] or \
            wave_rescaling[1] > wave_observations[1]:

        warnings.warn("Valid wavelength rescaling window must be between {0:8.2f} and {1:8.2f}".format(
            wave_observations[0], wave_observations[1]))

        return False
    else:
        return True


def _check_coadd_in_shared_data(shared_data, wavelength_range):
    if 'coadd' not in shared_data:
        shared_data['coadd'] = {
            'wavelength_range': wavelength_range[:]
        }
        shared_data['binned'] = {
            'wavelength_range': wavelength_range[:]
        }
    else:
        shared_data['coadd']['wavelength_range'][0] = min(shared_data['coadd']['wavelength_range'][0],
                                                          wavelength_range[0])

        shared_data['coadd']['wavelength_range'][1] = max(shared_data['coadd']['wavelength_range'][1],
                                                          wavelength_range[1])

        shared_data['binned']['wavelength_range'][0] = min(shared_data['binned']['wavelength_range'][0],
                                                           wavelength_range[0])

        shared_data['binned']['wavelength_range'][1] = max(shared_data['binned']['wavelength_range'][1],
                                                           wavelength_range[1])

    return shared_data


def prepare_datasets(config_in):
    """
    FITS files, telluric list etc. are retrieved at the beginning and converted to a pickle object
    to be processed to the next steps in the pipeline
    In this way all the changes performed on the fits files are preserved (sky correction, differential correction)
    """

    """ config_dictionary: dictionary with all the configuration parameters from config_in
        lists_dictionary: for each night, the list of files
    """

    night_dict = from_config_get_nights(config_in)

    instrument_dict = from_config_get_instrument(config_in)

    #try:
    #    shared_data = load_from_cpickle('shared', config_in['output'])
    #    loaded_shared_data = True
    #except:
    #    shared_data = {}
    #    loaded_shared_data = False

    if check_existence_cpickle('shared', config_in['output']):
        loaded_shared_data = True
    else:
        shared_data = {}
        loaded_shared_data = False

    pass_wavelength_rescaling = True
    pass_wavelength_master_out = True

    for night in night_dict:

        print('Processing data for night: ', night)
        print()

        """ List files are supposed to be in the same directory of the yaml file,
            NOT on the archive directory: in this way it is possible to try different
            combinations of nights and files without making a mess in the archive """
        files_list, files_transit_out, files_transit_in, files_transit_full, files_telluric, files_star_telluric = get_filelists(
            night_dict[night])

        """ fix the potential problem of multiple observations taken during the same night"""
        night_name = night_dict[night].get('night', night)

        lists_dictionary = {
            'observations': files_list,
            'star_telluric': files_star_telluric,
            'n_observations': len(files_list),
        }

        try:
            lists_dictionary['transit_out'] = files_transit_out
            lists_dictionary['transit_in'] = files_transit_in
            lists_dictionary['transit_full'] = files_transit_full
            lists_dictionary['n_transit_out'] = len(files_transit_out)
            lists_dictionary['n_transit_in'] = len(files_transit_in)
            lists_dictionary['n_transit_full'] = len(files_transit_full)
            lists_dictionary['telluric'] = files_telluric
            lists_dictionary['n_tellurics'] = len(files_telluric)
            write_transit_list = False
        except:
            print('  Input lists for transit in/out not found, proceeding to automatic selection and writing ')
            print()
            write_transit_list = True

        """ Retrieval on instrument characteristics """
        instrument = night_dict[night]['instrument']
        mask = night_dict[night].get('mask', 'undefined')
        archive_dir = instrument_dict[instrument]['data_archive']
        order_selection = instrument_dict[instrument]['orders']
        wavelength_rescaling = instrument_dict[instrument]['wavelength_rescaling']
        time_of_transit = night_dict[night]['time_of_transit']

        planet_dict = from_config_get_planet(config_in)
        star_dict = from_config_get_star(config_in)

        print("  # observations: ", lists_dictionary['n_observations'])
        try:
            print("  # out-transit obs: ", lists_dictionary['n_transit_out'])
            print("  # in-transit obs: ", lists_dictionary['n_transit_in'])
        except:
            pass
        print("  Instrument: ", instrument)
        print("  Archive DIR: ", archive_dir)
        print("  Night: ", night)
        print("  Mask:  ", mask)
        print("  WL rescaling: ", wavelength_rescaling)
        print()

        try:

            if not check_existence_cpickle('input_dataset_fibA', config_in['output'], night):
                raise ValueError()

            #pass_wavelength_rescaling = _check_wavelength_rescaling(wavelength_rescaling,
            #                                                        observations_A['coadd']['wavelength_range']
            #                                                        )

            #shared_data = _check_coadd_in_shared_data(shared_data,
            #                                          observations_A['coadd']['wavelength_range'])
            print("  Input data for night {0:s} successfully retrieved".format(night))

            if write_transit_list:
                observations_A = load_from_cpickle('input_dataset_fibA', config_in['output'], night)

                lists_dictionary = _write_transit_list(observations_A,
                                                       lists_dictionary,
                                                       night_dict[night],
                                                       planet_dict)

                save_to_cpickle('lists', lists_dictionary, config_in['output'], night)
                print("  List rewritten, be careful however that you may incur in extra problems if they have changed")

            try:
                observational_parameters = load_from_cpickle('observational_pams', config_in['output'], night)
            except:
                observations_A = load_from_cpickle('input_dataset_fibA', config_in['output'], night)
                observational_parameters = _get_observational_parameters(observations_A,
                                                                         lists_dictionary,
                                                                         night_dict[night],
                                                                         instrument_dict[instrument],
                                                                         star_dict,
                                                                         planet_dict)
                save_to_cpickle('observational_pams', observational_parameters, config_in['output'], night)
                print("  New observational parameters loaded successfully")

            for key_name, key_val in observational_parameters['RV_star'].items():
                print("   RV star  {0:s}: {1}".format(key_name, key_val))
            print()
            continue

        except ValueError:
            pass

        observations_A = {}
        observations_s1d_A = {}
        observations_B = {}
        calib_data_A = {}
        calib_data_B = {}

        for obs in lists_dictionary['observations']:

            print("  Reading ", obs, " associated files")

            observations_A[obs], observations_s1d_A[obs] = \
                get_input_data(instrument, night_dict[night], archive_dir + night_name, obs, mask, skip_s1d=False,
                               order_selection=order_selection)   

            #""" Zero or negative values are identified, flagged and substituted with another value """
            #replacement = 0.01
            #observations_A[obs]['null'] = (observations_A[obs]['e2ds'] <= replacement)
            #observations_A[obs]['e2ds'][observations_A[obs]['null']] = replacement

            """ Negative values are just statistical noise around the null flux points,
            removing them would bias the flux level of the sky towards
            higher values

            We proceed in this way:
            -  identify the negative values and make a statistics of their
                average value
            - if in relevant number, we assume that the median of their absolute
                values corresponds to the noise floor
            - we add the noise floor to the error estimate 
            """

            observations_A[obs]['null'] = (observations_A[obs]['e2ds'] <= 0.0)
            if (np.sum(observations_A[obs]['null']) > 30):
                observations_A[obs]['noise_floor'] = np.median(np.abs(
                    observations_A[obs]['e2ds'][observations_A[obs]['null']]))
            else:
                if observations_A[obs].get('absolute_flux', True):
                    observations_A[obs]['noise_floor'] = 1.0000
                else:
                    observations_A[obs]['noise_floor'] = 0.00001


            observations_A[obs]['e2ds_err'] = np.sqrt(observations_A[obs]['e2ds_err']**2 + observations_A[obs]['noise_floor']**2)
            #observations_A[obs]['e2ds_err'] = np.sqrt(observations_A[obs]['e2ds'])

            if 'n_orders' not in observations_A or 'n_pixels' not in observations_A:
                observations_A['n_orders'] = observations_A[obs]['n_orders']
                observations_A['n_pixels'] = observations_A[obs]['n_pixels']
                calib_data_A = get_calib_data(instrument, night_dict[night], archive_dir + night_name, obs,
                                              order_selection=order_selection)

            """ Updating info on shared data """
            if 'coadd' not in observations_A:
                observations_A['coadd'] = {
                    'wavelength_range': [np.min(observations_A[obs]['wave'][0, :]),
                                         np.max(observations_A[obs]['wave'][-1, :])]
                }

            else:
                observations_A['coadd']['wavelength_range'][0] = min(observations_A['coadd']['wavelength_range'][0],
                                                                     np.min(observations_A[obs]['wave'][0, :]))

                observations_A['coadd']['wavelength_range'][1] = max(observations_A['coadd']['wavelength_range'][1],
                                                                     np.max(observations_A[obs]['wave'][-1, :]))

            """ Reading the fiber B counter part
                If the target has been observed in ThAr or FP mode, fiber B data will not be accessible
            """
            has_fiber_B = False
            try:

                observations_B[obs], _ = get_input_data(instrument, night_dict[night], archive_dir + night_name, obs, mask,
                                                        fiber='B', order_selection=order_selection)

                """ Negative values are just statistical noise around the null flux points,
                removing them would bias the flux level of the sky towards
                higher values

                We proceed in this way:
                -   identify the negative values and make a statistics of their
                    average value
                - if in relevant number, we assume that the their median value
                  corresponds to the noise floor
                - we add the noise floor to the error estimate
                """

                #""" Zero or negative values are identified, flagged and substituted with another value """
                #replacement = 0.01
                observations_B[obs]['null'] = (observations_B[obs]['e2ds'] <= 0.0)
                if (np.sum(observations_B[obs]['null']) > 30):
                    observations_B[obs]['noise_floor'] = np.median(np.abs(
                        observations_B[obs]['e2ds'][observations_B[obs]['null']]))
                else:
                    if observations_B[obs].get('absolute_flux', True):
                        observations_B[obs]['noise_floor'] = 1.0000
                    else:
                        observations_B[obs]['noise_floor'] = 0.00001                #observations_B[obs]['e2ds'][observations_B[obs]['null']] = replacement

                observations_B[obs]['e2ds_err'] = np.sqrt(np.abs(observations_B[obs]['e2ds'])) + observations_B[obs]['noise_floor']

                if 'n_orders' not in observations_B or 'n_pixels' not in observations_B:
                    observations_B['n_orders'] = observations_B[obs]['n_orders']
                    observations_B['n_pixels'] = observations_B[obs]['n_pixels']

                    calib_data_B = get_calib_data(instrument, night_dict[night], archive_dir + night_name, obs,
                                                  fiber='B', order_selection=order_selection)
                has_fiber_B = True
            except:
                pass

        """ Building the base (array of wavelengths) for coadded spectra within the same night """

        observations_A['coadd']['wave'] = np.arange(observations_A['coadd']['wavelength_range'][0],
                                                    observations_A['coadd']['wavelength_range'][1],
                                                    instrument_dict[instrument]['wavelength_step'], dtype=np.double)

        observations_A['coadd']['size'] = np.size(observations_A['coadd']['wave'])
        observations_A['coadd']['step'] = np.ones(observations_A['coadd']['size'], dtype=np.double) * \
                                          instrument_dict[instrument]['wavelength_step']
        print()
        print("  Fixing the observation lists if they are missing")
        if write_transit_list:
            lists_dictionary = _write_transit_list(observations_A,
                                                   lists_dictionary,
                                                   night_dict[night],
                                                   planet_dict)

        print()
        print("  Computing the RV shift outside the transit, and store it to an additional file for quick access ")
        observational_parameters = _get_observational_parameters(observations_A,
                                                                 lists_dictionary,
                                                                 night_dict[night],
                                                                 instrument_dict[instrument],
                                                                 star_dict,
                                                                 planet_dict)
        print()
        for key_name, key_val in observational_parameters['RV_star'].items():
            print(" RV star  {0:s}: {1:f}".format(key_name, key_val))

        print()
        print("  Writing dataset files for night ", night, "  fiber A")
        save_to_cpickle('lists', lists_dictionary, config_in['output'], night)
        save_to_cpickle('input_dataset_fibA', observations_A, config_in['output'], night)
        save_to_cpickle('input_dataset_s1d_fibA', observations_s1d_A, config_in['output'], night)
        save_to_cpickle('calibration_fibA', calib_data_A, config_in['output'], night)
        save_to_cpickle('observational_pams', observational_parameters, config_in['output'], night)

        if has_fiber_B:
            observations_B['coadd'] = {}
            observations_B['coadd']['wave'] = observations_A['coadd']['wave'].copy()
            observations_B['coadd']['size'] = np.size(observations_B['coadd']['wave'])
            observations_B['coadd']['step'] = np.ones(observations_B['coadd']['size'], dtype=np.double) * \
                                              instrument_dict[instrument]['wavelength_step']

            print("  Writing dataset files for night ", night, "  fiber B")
            save_to_cpickle('input_dataset_fibB', observations_B, config_in['output'], night)
            save_to_cpickle('calibration_fibB', calib_data_B, config_in['output'], night)

        """ Running some checks to see if input parameters have been configured properly """
        wavelength_rescaling = instrument_dict[instrument]['wavelength_rescaling']

        pass_wavelength_rescaling = _check_wavelength_rescaling(
            wavelength_rescaling,
            observations_A['coadd']['wavelength_range']
        )
        print()

        if loaded_shared_data: continue

        """ Setting up the base arrays for all the nights"""

        shared_data = _check_coadd_in_shared_data(shared_data, observations_A['coadd']['wavelength_range'])

    if loaded_shared_data: return

    """ Building the base (array of wavelengths) for master-out and coadd spectra
        We do it now to be sure that it will be the same for the whole pipeline
    """
    print("  Creating the shared arrays")
    print()

    shared_data['coadd']['wavelength_range'][0] += 2.0
    shared_data['coadd']['wavelength_range'][1] -= 2.0

    shared_data['coadd']['wave'] = np.arange(shared_data['coadd']['wavelength_range'][0],
                                             shared_data['coadd']['wavelength_range'][1],
                                             config_in['instruments']['shared']['wavelength_step'], dtype=np.double)
    shared_data['coadd']['size'] = np.size(shared_data['coadd']['wave'])
    shared_data['coadd']['step'] = np.ones(shared_data['coadd']['size'], dtype=np.double) * \
                                   config_in['instruments']['shared']['wavelength_step']

    shared_data['binned']['wave'] = np.arange(shared_data['coadd']['wavelength_range'][0],
                                              shared_data['coadd']['wavelength_range'][1],
                                              config_in['instruments']['shared']['wavelength_step']
                                              * config_in['master-out']['binning_factor'], dtype=np.double)
    shared_data['binned']['size'] = np.size(shared_data['binned']['wave'])
    shared_data['binned']['step'] = np.ones(shared_data['binned']['size'], dtype=np.double) * \
                                    config_in['instruments']['shared']['wavelength_step'] * \
                                    config_in['master-out']['binning_factor']

    if config_in['master-out']['wavelength_range'][0] < shared_data['coadd']['wavelength_range'][0] or \
            config_in['master-out']['wavelength_range'][1] > shared_data['coadd']['wavelength_range'][1]:
        warnings.warn("ERROR: Valid master_out wavelength window must be between {0:8.2f} and {1:8.2f}".format(
            shared_data['coadd']['wavelength_range'][0],
            shared_data['coadd']['wavelength_range'][1]))

        pass_wavelength_master_out = False

    shared_data['master-out'] = {}
    shared_data['master-out']['wave'] = np.arange(config_in['master-out']['wavelength_range'][0],
                                                  config_in['master-out']['wavelength_range'][1],
                                                  config_in['master-out']['wavelength_step'], dtype=np.double)

    shared_data['master-out']['size'] = np.size(shared_data['master-out']['wave'])
    shared_data['master-out']['step'] = np.ones(shared_data['master-out']['size'], dtype=np.double) * \
                                        config_in['master-out']['wavelength_step']

    if not (pass_wavelength_master_out and pass_wavelength_rescaling):
        raise ValueError("ERROR: check the previous warnings to see where you are doing it worng")

    print("  COADD wavelength range between {0:8.2f} and {1:8.2f}".format(
        shared_data['coadd']['wavelength_range'][0], shared_data['coadd']['wavelength_range'][1]))
    print("  COADD wavelength step: {0:5.3f}".format(config_in['instruments']['shared']['wavelength_step']))
    print()
    print("Saving shared data")
    save_to_cpickle('shared', shared_data, config_in['output'])
    print()


def _write_transit_list(observations_A, lists_dict, night_dict_key, planet_dict):
    fileout_transit_in_list = open(night_dict_key['in_transit'], 'w')
    fileout_transit_out_list = open(night_dict_key['out_transit'], 'w')
    fileout_transit_full_list = open(night_dict_key['full_transit'], 'w')

    try:
        total_transit_start = np.atleast_1d(night_dict_key['time_of_transit'])[0] - np.atleast_1d(planet_dict['total_transit_duration'])[0] / 2.
        total_transit_end = np.atleast_1d(night_dict_key['time_of_transit'])[0] + np.atleast_1d(planet_dict['total_transit_duration'])[0] / 2.
        full_transit_start = np.atleast_1d(night_dict_key['time_of_transit'])[0] - np.atleast_1d(planet_dict['full_transit_duration'])[0] / 2.
        full_transit_end = np.atleast_1d(night_dict_key['time_of_transit'])[0] + np.atleast_1d(planet_dict['full_transit_duration'])[0] / 2.
    except KeyError:
        total_transit_start = np.atleast_1d(night_dict_key['time_of_transit'])[0] - np.atleast_1d(planet_dict['transit_duration'])[0] / 2.
        total_transit_end = np.atleast_1d(night_dict_key['time_of_transit'])[0] + np.atleast_1d(planet_dict['transit_duration'])[0] / 2.
        full_transit_start = np.atleast_1d(night_dict_key['time_of_transit'])[0] - np.atleast_1d(planet_dict['transit_duration'])[0] / 2.
        full_transit_end = np.atleast_1d(night_dict_key['time_of_transit'])[0] + np.atleast_1d(planet_dict['transit_duration'])[0] / 2.
        print('*** unclear transit duration, ingress/egress observations will be considered full-transit')

    for obs in lists_dict['observations']:

        """ Check if the file should be in transit_in or transit_out list, in case they are not present"""
        #phase_internal = (observations_A[obs]['BJD'] - np.atleast_1d(night_dict_key['time_of_transit'])[0]) / \
        #                 np.atleast_1d(planet_dict['period'])[0]
        #if np.abs(phase_internal) <= planet_dict['transit_duration'][0] / 2. / planet_dict['period'][0]:
        #    fileout_transit_in_list.write('{0:s}\n'.format(obs))
        #else:
        #    fileout_transit_out_list.write('{0:s}\n'.format(obs))

        exptime_seconds = observations_A[obs]['EXPTIME'] / 86400.
        """BJD times have been already corrected to match mid-exposure epochs  """

        if observations_A[obs]['BJD'] + exptime_seconds/2. < total_transit_start \
            or observations_A[obs]['BJD'] - exptime_seconds/2. > total_transit_end:
            fileout_transit_out_list.write('{0:s}\n'.format(obs))
        else:
            fileout_transit_in_list.write('{0:s}\n'.format(obs))
        if observations_A[obs]['BJD'] - exptime_seconds/2. > full_transit_start \
            and observations_A[obs]['BJD'] + exptime_seconds/2. < full_transit_end:
            fileout_transit_full_list.write('{0:s}\n'.format(obs))

    fileout_transit_in_list.close()
    fileout_transit_out_list.close()
    fileout_transit_full_list.close()
    files_list, files_transit_out, files_transit_in, files_transit_full, files_telluric, files_star_telluric = get_filelists(night_dict_key)

    lists_dict['transit_out'] = files_transit_out
    lists_dict['transit_in'] = files_transit_in
    lists_dict['transit_full'] = files_transit_full
    lists_dict['n_transit_out'] = np.size(files_transit_out)
    lists_dict['n_transit_in'] = np.size(files_transit_in)
    lists_dict['n_transit_full'] = np.size(files_transit_full)


    try:
        lists_dict['telluric'] = files_telluric
        lists_dict['n_tellurics'] = np.size(files_telluric)
    except:
        lists_dict['telluric'] = files_transit_out.copy()
        lists_dict['n_tellurics'] = np.size(files_transit_out)

    print(" # observations: ", lists_dict['n_observations'])
    print(" # out-transit obs: ", lists_dict['n_transit_out'])
    print(" # in-transit obs: ", lists_dict['n_transit_in'])
    print(" # full-transit obs: ", lists_dict['n_transit_full'])

    return lists_dict


def _get_observational_parameters(observations_A, lists_dict, night_dict_key, instrument_dict_key, star_dict, planet_dict):

    observational_parameters = {
        'instrument': night_dict_key['instrument'],
        'mask': night_dict_key['mask'],
        'archive_dir': instrument_dict_key['data_archive'],
        'wavelength_rescaling': instrument_dict_key['wavelength_rescaling'],
        'time_of_transit': np.atleast_1d(night_dict_key['time_of_transit'])[0],
        'refraction_method': instrument_dict_key['refraction']['method'],
        'refraction_fit_order': instrument_dict_key['refraction']['fit_order'],
        'refraction_fit_iters': instrument_dict_key['refraction']['fit_iters'],
        'refraction_fit_sigma': instrument_dict_key['refraction']['fit_sigma'],
        'refraction_knots_spacing': instrument_dict_key['refraction']['knots_spacing'],
        'linear_fit_method': instrument_dict_key['linear_fit_method'],
        'n_orders': observations_A['n_orders'],
        'n_pixels': observations_A['n_pixels'],
        'RV_star': {}
    }

    rv_out = []
    bjd0_out = []

    for obs in lists_dict['transit_out']:
        bjd0_out.extend([observations_A[obs]['BJD'] - observational_parameters['time_of_transit']])
        rv_out.extend([observations_A[obs]['RVC']])

    observational_parameters['RV_star']['slope'], \
    observational_parameters['RV_star']['intercept'], \
    observational_parameters['RV_star']['r_value'], \
    observational_parameters['RV_star']['p_value'], \
    observational_parameters['RV_star']['std_err'] = sci_stats.linregress(bjd0_out, rv_out)

    berv_list = []
    rvc_stack = np.zeros(len(lists_dict['observations']))
    for i, obs in enumerate(lists_dict['observations']):
        berv_list.extend([observations_A[obs]['BERV']])
        observational_parameters[obs] = {
            'BJD': observations_A[obs]['BJD'],
            'mBJD': observations_A[obs]['BJD']-2450000.0000,
            'RVC': observations_A[obs]['RVC'],
            'AIRMASS': observations_A[obs]['AIRMASS'],
            'BERV': observations_A[obs]['BERV'],
            'EXPTIME': observations_A[obs]['EXPTIME']
        }
        rvc_stack[i] = observations_A[obs]['RVC']

    observational_parameters['BERV_avg'] = np.average(berv_list)

    observational_parameters['RV_star']['RV_from_CCF'] = False
    observational_parameters['RV_star']['RV_from_analytical_solution'] = False


    if night_dict_key['use_rv_from_ccf']:
        observational_parameters['RV_star']['RV_from_CCF'] = True
        rvc_systemic = np.average(rvc_stack)
    elif night_dict_key['use_analytical_rvs']:
        observational_parameters['RV_star']['RV_from_analytical_solution'] = True
        rvc_systemic = star_dict['RV_gamma'][0]
        observational_parameters['RV_star']['RV_semiamplitude'] = star_dict['RV_semiamplitude'][0]
    else:
        rvc_systemic = observational_parameters['RV_star']['intercept']

    observational_parameters['RV_star']['RV_systemic'] = rvc_systemic

    for obs in lists_dict['observations']:

        if night_dict_key['use_rv_from_ccf']:
            rvc_bjdshift = observations_A[obs]['RVC']
        elif night_dict_key['use_analytical_rvs']:
            rvc_bjdshift = - observational_parameters['RV_star']['RV_semiamplitude'] * np.sin(2 * np.pi * \
                           (observational_parameters[obs]['BJD'] - observational_parameters['time_of_transit'])/planet_dict['period'][0])
        else:
            rvc_bjdshift = observational_parameters['RV_star']['slope'] * \
                           (observational_parameters[obs]['BJD'] - observational_parameters['time_of_transit'])

        observational_parameters[obs]['RV_bjdshift'] = rvc_bjdshift
        observational_parameters[obs]['rv_shift_ORF2SRF'] = observational_parameters[obs]['BERV'] - \
                                                            (rvc_systemic + rvc_bjdshift)

        """ Old definition
        observational_parameters[obs]['rv_shift_ORF2SRF'] = observational_parameters[obs]['BERV'] - \
                                                            (observational_parameters['RV_star']['intercept'] +
                                                             observational_parameters['RV_star']['slope'] *
                                                             (observational_parameters[obs][
                                                                  'BJD'] - time_of_transit))
        """

        """ Slight modification of the RV shift to minimize the rebinning error at the wings of the spectra
        BRF = Solar System Barycentric Reference frame
            rv_shift_ORF2BRF = rv_shift_ORF2SRF_mod + rv_shift_ORF2SRF_res
        """
        observational_parameters[obs]['rv_shift_ORF2BRF'] = \
            observational_parameters[obs]['BERV']

        observational_parameters[obs]['rv_shift_ORF2BRF_mod'] = \
            observational_parameters[obs]['BERV'] - observational_parameters['BERV_avg']

        """ Slight modification of the RV shift to minimize the rebinning error at the wings of the spectra
            rv_shift_ORF2SRF = rv_shift_ORF2SRF_mod + rv_shift_ORF2SRF_res
        """
        observational_parameters[obs]['rv_shift_ORF2SRF_mod'] = \
            observational_parameters[obs]['BERV'] - observational_parameters['BERV_avg'] - rvc_bjdshift

        observational_parameters[obs]['rv_shift_ORF2SRF_res'] = \
            observational_parameters['BERV_avg'] - rvc_systemic

        """ RV shift from the observer RF to the planet RF
            STRONG ASSUMPTIONS:
                - there is only the transiting planet in the system
                - the planet has null eccentricity
                - linear approximation or the orbit near the transit event

            Computation is performed by moving to the Solar Barycenter, than to the Stellar System Barycenter
            and finally onto the planet
        """
        observational_parameters[obs]['rv_shift_ORF2PRF'] = \
            observational_parameters[obs]['BERV'] \
            - rvc_systemic \
            - planet_dict['RV_semiamplitude'][0] \
            * (observational_parameters[obs]['BJD'] - observational_parameters['time_of_transit']) \
            / planet_dict['period'][0] * 2 * np.pi

        """ RV shift from Stellar Rest Frame to Planetary Rest Frame
            We have to take into account the RV of star relatively to the Barycenter
        """
        observational_parameters[obs]['rv_shift_SRF2PRF'] = \
            + rvc_bjdshift \
            - planet_dict['RV_semiamplitude'][0] \
            * (observational_parameters[obs]['BJD'] - observational_parameters['time_of_transit']) \
            / planet_dict['period'][0] * 2 * np.pi

    observational_parameters['rv_shift_ORF2SRF_res'] = \
        observational_parameters['BERV_avg'] - rvc_systemic

    return observational_parameters


def plot_dataset(config_in, night_input=''):
    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_dict:

        print()
        print("Plotting dataset                           Night: ", night)

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        instrument = night_dict[night]['instrument']
        wavelength_rescaling = instrument_dict[instrument]['wavelength_rescaling']

        """ Retrieving the observations"""
        input_data = load_from_cpickle('input_dataset_fibA', config_in['output'], night)

        """ Retrieving the calibration data """
        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)

        colors, cmap, line_colors = make_color_array(lists, input_data)
        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])
        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)

        cbax1 = plt.subplot(gs[:, 1])


        for i, obs in enumerate(lists['observations']):

            rescaling = compute_rescaling(input_data[obs]['wave'], input_data[obs]['e2ds'], wavelength_rescaling)
            for order in range(0,input_data[obs]['n_orders']):


                ax1.plot(input_data[obs]['wave'][order,:], input_data[obs]['e2ds'][order,:]/ rescaling, zorder=i, lw=1,
                        c=line_colors[i], alpha=0.5)
                ax2.plot(input_data[obs]['wave'][order,:], input_data[obs]['e2ds'][order,:] / rescaling, zorder=-i, lw=1,
                        c=line_colors[i], alpha=0.5)

        ax1.legend(loc=3)
        ax1.set_title('Night: ' + night)

        ax2.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=colors[0], vmax=colors[-1]))
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()



from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.transmission_spectrum import *
from SLOPpy.transmission_spectrum_average import *

#__all__ = ['compute_transmission_spectrum_planetRF_iterative',
#           'plot_transmission_spectrum_planetRF_iterative',
#           'compute_transmission_spectrum_stellarRF_iterative',
#           'plot_transmission_spectrum_stellarRF_iterative',
#           'compute_transmission_spectrum_observerRF_iterative',
#           'plot_transmission_spectrum_observerRF_iterative',
#           'compute_transmission_spectrum_iterative',
#           'plot_transmission_spectrum_iterative']


def compute_transmission_spectrum_planetRF(config_in, lines_label):
    compute_transmission_spectrum(config_in, lines_label, reference='planetRF')


def plot_transmission_spectrum_planetRF(config_in, lines_label, night_input, results_input=''):
    plot_transmission_spectrum(config_in, lines_label, night_input, results_input, reference='planetRF')


def compute_transmission_spectrum_stellarRF(config_in, lines_label):
    compute_transmission_spectrum(config_in, lines_label, reference='stellarRF')


def plot_transmission_spectrum_stellarRF(config_in, lines_label, night_input, results_input=''):
    plot_transmission_spectrum(config_in, lines_label, night_input, results_input, reference='stellarRF')


def compute_transmission_spectrum_observerRF(config_in, lines_label):
    compute_transmission_spectrum(config_in, lines_label, reference='observerRF')


def plot_transmission_spectrum_observerRF(config_in, lines_label, night_input, results_input=''):
    plot_transmission_spectrum(config_in, lines_label, night_input, results_input, reference='observerRF')






def compute_transmission_spectrum_planetRF_iterative(config_in, lines_label):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        compute_transmission_spectrum(config_in, lines_label, reference='planetRF', pca_iteration=it)

def compute_transmission_spectrum_stellarRF_iterative(config_in, lines_label):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        compute_transmission_spectrum(config_in, lines_label, reference='stellarRF', pca_iteration=it)

def compute_transmission_spectrum_observerRF_iterative(config_in, lines_label):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        compute_transmission_spectrum(config_in, lines_label, reference='observerRF', pca_iteration=it)

def compute_transmission_spectrum_iterative(config_in, lines_label):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        compute_transmission_spectrum(config_in, lines_label, reference='planetRF', pca_iteration=it)


def  plot_transmission_spectrum_planetRF_iterative(config_in, lines_label, night_input, results_input=''):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        plot_transmission_spectrum(config_in, lines_label, night_input, results_input, reference='planetRF', pca_iteration=it)

def plot_transmission_spectrum_stellarRF_iterative(config_in, lines_label, night_input, results_input=''):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        plot_transmission_spectrum(config_in, lines_label, night_input, results_input, reference='stellarRF', pca_iteration=it)

def plot_transmission_spectrum_observerRF_iterative(config_in, lines_label, night_input, results_input=''):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        plot_transmission_spectrum(config_in, lines_label, night_input, results_input, reference='observerRF', pca_iteration=it)

def plot_transmission_spectrum_iterative(config_in, lines_label, night_input, results_input=''):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        plot_transmission_spectrum(config_in, lines_label, night_input, results_input, reference='planetRF', pca_iteration=it)




def compute_transmission_spectrum_average_planetRF(config_in, lines_label):
    compute_transmission_spectrum_average(config_in, lines_label, reference='planetRF')


def compute_transmission_spectrum_average_observerRF(config_in, lines_label):
    compute_transmission_spectrum_average(config_in, lines_label, reference='observerRF')


def compute_transmission_spectrum_average_stellarRF(config_in, lines_label):
    compute_transmission_spectrum_average(config_in, lines_label, reference='stellarRF')


def plot_transmission_spectrum_average_planetRF(config_in, lines_label, night_input='', results_input=''):
    plot_transmission_spectrum_average(config_in, lines_label, night_input, results_input, reference='planetRF')


def plot_transmission_spectrum_average_observerRF(config_in, lines_label, night_input='', results_input=''):
    plot_transmission_spectrum_average(config_in, lines_label, night_input, results_input, reference='observerRF')


def plot_transmission_spectrum_average_stellarRF(config_in, lines_label, night_input='', results_input=''):
    plot_transmission_spectrum_average(config_in, lines_label, night_input, results_input, reference='stellarRF')







def compute_transmission_spectrum_average_planetRF_iterative(config_in, lines_label):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        compute_transmission_spectrum_average(config_in, lines_label, reference='planetRF', pca_iteration=it)


def compute_transmission_spectrum_average_observerRF_iterative(config_in, lines_label):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        compute_transmission_spectrum_average(config_in, lines_label, reference='observerRF', pca_iteration=it)


def compute_transmission_spectrum_average_stellarRF_iterative(config_in, lines_label):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        compute_transmission_spectrum_average(config_in, lines_label, reference='stellarRF', pca_iteration=it)


def compute_transmission_spectrum_average_iterative(config_in, lines_label):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        compute_transmission_spectrum_average(config_in, lines_label, reference='planetRF', pca_iteration=it)


def plot_transmission_spectrum_average_planetRF_iterative(config_in, lines_label, night_input='', results_input=''):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        plot_transmission_spectrum_average(config_in, lines_label, night_input, results_input, reference='planetRF', pca_iteration=it)


def plot_transmission_spectrum_average_observerRF_iterative(config_in, lines_label, night_input='', results_input=''):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        plot_transmission_spectrum_average(config_in, lines_label, night_input, results_input, reference='observerRF', pca_iteration=it)


def plot_transmission_spectrum_average_stellarRF_iterative(config_in, lines_label, night_input='', results_input=''):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        plot_transmission_spectrum_average(config_in, lines_label, night_input, results_input, reference='stellarRF', pca_iteration=it)


def plot_transmission_spectrum_average_iterative(config_in, lines_label, night_input='', results_input=''):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        plot_transmission_spectrum_average(config_in, lines_label, night_input, results_input, reference='planetRF', pca_iteration=it)



from __future__ import print_function, division
import numpy as np
from astropy.io import fits
from SLOPpy.subroutines.rebin_subroutines import *
from SLOPpy.subroutines.constants import *
from SLOPpy.instruments.HARPN_DRSv3 import *
from SLOPpy.instruments.HARPS_DRSv3 import *
from SLOPpy.instruments.PEPSI_reduced import *
from SLOPpy.instruments.ESPRESSO import *

def get_calib_data(instrument, night_dict, archive, file_rad, fiber='A', order_selection=None):

    if instrument =='HARPS-N':
        return HARPN_DRSv3_get_calib_data(archive, file_rad, fiber=fiber, order_selection=order_selection)
    elif instrument =='HARPS':
        return HARPS_DRSv3_get_calib_data(archive, file_rad, fiber=fiber, order_selection=order_selection)
    elif instrument in ['PEPSI', 'PEPSI_red', 'PEPSI_blue']:
        return PEPSI_get_calib_data(archive, file_rad, fiber=fiber, order_selection=order_selection)
    elif instrument in ['ESPRESSO', 'ESPRESSO_odd', 'ESPRESSO_even']:
        return ESPRESSO_get_calib_data(archive, file_rad, night_dict, fiber=fiber, order_selection=order_selection)
    else:
        raise ValueError("Instrument not supported")


def get_input_data(instrument, night_dict, archive, file_rad, mask, fiber='A', skip_ccf=None, skip_s1d=True, order_selection=None):

    if instrument =='HARPS-N':
        return HARPN_DRSv3_get_input_data(archive, file_rad, mask, fiber=fiber, skip_ccf=skip_ccf, skip_s1d=skip_s1d, order_selection=order_selection)
    elif instrument =='HARPS':
        return HARPS_DRSv3_get_input_data(archive, file_rad, mask, fiber=fiber, skip_ccf=skip_ccf, skip_s1d=skip_s1d, order_selection=order_selection)
    elif instrument in ['PEPSI', 'PEPSI_red', 'PEPSI_blue']:
        return PEPSI_get_input_data(archive, file_rad, mask, fiber=fiber, skip_ccf=skip_ccf, skip_s1d=skip_s1d, order_selection=order_selection)
    elif instrument in ['ESPRESSO', 'ESPRESSO_odd', 'ESPRESSO_even']:
        return ESPRESSO_get_input_data(archive, file_rad, night_dict, fiber=fiber, skip_ccf=skip_ccf, skip_s1d=skip_s1d, order_selection=order_selection)
    
        """ Keywords in night:
        telescope: UT1
        spectral_selection: odd % even
        apply_ESO_telluric_correction: True
        use_ESO_sky_correction: True 
        use_ESO_deblazed: True
        """
    
    else:
        raise ValueError("Instrument not supported")



from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.shortcuts import *
from SLOPpy.subroutines.plot_subroutines import *

__all__ = ["compute_differential_refraction_preparation"]

def compute_differential_refraction_preparation(config_in, append_name=None):

    if append_name:
        subroutine_name = 'differential_refraction_preparation_' + append_name
        filename = 'refraction_'+append_name
    else:
        subroutine_name = 'differential_refraction_preparation'
        filename = 'refraction'

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)

    shared_data = load_from_cpickle('shared', config_in['output'])
    reference_flux = np.empty([len(night_dict), shared_data['coadd']['size']], dtype=np.double)
    reference_wght = np.zeros([len(night_dict), shared_data['coadd']['size']], dtype=np.double)
    reference_mask = np.ones([len(night_dict), shared_data['coadd']['size']], dtype=bool)
    compute_reference = False

    """ make sure that all the spectra are computed in the same reference system if cross-calibrations is used"""
    absolute_SRF = False
    for n_night, night in enumerate(night_dict):
        if night_dict[night]['refraction'].get('reference_night', False) \
                or night_dict[night]['refraction'].get('reference_instrument', False):
            absolute_SRF = True

    for n_night, night in enumerate(night_dict):

        try:
            preparation = load_from_cpickle(filename +'_preparation', config_in['output'], night)
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Retrieved'))
            continue
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Computing'))
            print()

        instrument = night_dict[night]['instrument']

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'],
                                           use_refraction=False)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        defined_reference = night_dict[night]['refraction'].get('reference', False)

        if defined_reference:
            preparation = {
                'subroutine': 'differential_refraction_preparation',
                'coadd': {
                    'wave': shared_data['coadd']['wave'],
                    'step': shared_data['coadd']['step'],
                    'size': shared_data['coadd']['size']
                },
                'absolute_SRF': True,
                'reference_coadd': True
            }
        else:
            preparation = {
                'subroutine': 'differential_refraction_preparation',
                'coadd': {
                    'wave': input_data['coadd']['wave'],
                    'step': input_data['coadd']['step'],
                    'size': input_data['coadd']['size'],
                },
                'absolute_SRF': absolute_SRF,
                'reference_coadd': False
            }

        total_flux = np.empty([len(lists['observations']), preparation['coadd']['size']], dtype=np.double)
        total_wght = np.zeros([len(lists['observations']), preparation['coadd']['size']], dtype=np.double)
        total_mask = np.ones([len(lists['observations']), preparation['coadd']['size']], dtype=bool)

        """ Rebinning of all the spectra """
        for n_obs, obs in enumerate(lists['observations']):

            print("  Spectral rebinning - Processing: ", obs)
            preparation[obs] = {}

            """ Rebinning of the spectra in the SRF, except for a fixed constant in order to minimize
                the difference between  
            """
            if preparation['absolute_SRF']:
                rv_shift = observational_pams[obs]['rv_shift_ORF2SRF']
            else:
                rv_shift = observational_pams[obs]['rv_shift_ORF2SRF_mod']

            preserve_flux = input_data[obs].get('absolute_flux', True)

            preparation[obs]['flux_rebinned_stellarRF'] = \
                rebin_2d_to_1d(input_data[obs]['wave'],
                               input_data[obs]['step'],
                               input_data[obs]['e2ds'],
                               calib_data['blaze'],
                               preparation['coadd']['wave'],
                               preparation['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=rv_shift)

            err_flux_rebinned_SRF = \
                rebin_2d_to_1d(input_data[obs]['wave'],
                               input_data[obs]['step'],
                               input_data[obs]['e2ds_err'],
                               calib_data['blaze'],
                               preparation['coadd']['wave'],
                               preparation['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=rv_shift,
                               is_error=True)

            """ Zero or negative values are identified, flagged and substituted with another value """
            #processed[obs]['flux_rebinned_stellarRF'], \
            #processed[obs]['err_flux_rebinned_SRF'], \
            #processed[obs]['flux_rebinned_SRF_null'] = \
            #    replace_values_errors_with_interpolation_1d(processed[obs]['flux_rebinned_stellarRF'],
            #                                                processed[obs]['err_flux_rebinned_SRF'],
            #                                                force_positive=True)

            #processed[obs]['rescaling'], processed[obs]['rescaled'], processed[obs]['rescaled_err'] = \
            #    perform_rescaling(processed['coadd']['wave'],
            #                      processed[obs]['flux_rebinned_stellarRF'],
            #                      processed[obs]['err_flux_rebinned_SRF'],
            #                      observational_pams['wavelength_rescaling'])
            #

            """ Zero or negative values are identified, flagged and substituted
            with another value """

            preparation[obs]['flux_rebinned_stellarRF'], \
            err_flux_rebinned_SRF, \
            flux_rebinned_SRF_null = \
                replace_values_errors_with_interpolation_1d(preparation[obs]['flux_rebinned_stellarRF'],
                                                            err_flux_rebinned_SRF,
                                                            force_positive=True)

            preparation[obs]['rescaling'], preparation[obs]['rescaled'], preparation[obs]['rescaled_err'] = \
                perform_rescaling(preparation['coadd']['wave'],
                                  preparation[obs]['flux_rebinned_stellarRF'],
                                  err_flux_rebinned_SRF,
                                  observational_pams['wavelength_rescaling'])

            #if instrument_dict[instrument]['refraction'].get('reference_night', False):
            #    if night != instrument_dict[instrument]['refraction']['reference_night']:
            #        continue
            #
            #if instrument_dict[instrument]['refraction'].get('reference_instrument', False):
            #    if instrument != instrument_dict[instrument]['refraction']['reference_instrument']:
            #        continue

            if night_dict[night]['refraction'].get('use_all_observations', False) or obs in lists['telluric']:
                total_flux[n_obs, :] = preparation[obs]['rescaled']
                total_mask[n_obs, :] = flux_rebinned_SRF_null
                total_wght[n_obs, :] = 1. / (preparation[obs]['rescaled_err'] ** 2)
                print("      Observation added to reference spectrum")

        masked_array = np.ma.array(total_flux, mask=total_mask)
        rescaled_mask, sum_weights = np.ma.average(masked_array,
                                                   weights=total_wght,
                                                   axis=0,
                                                   returned=True)

        preparation['coadd']['rescaled'] = rescaled_mask.filled(0.00)
        sum_weights[sum_weights <= 0.0] = 1.0
        preparation['coadd']['rescaled_err'] = 1. / np.sqrt(sum_weights)

        preparation['coadd']['rescaled'], preparation['coadd']['rescaled_err'], preparation['coadd']['null'] = \
            replace_values_errors_with_interpolation_1d(preparation['coadd']['rescaled'],
                                                        preparation['coadd']['rescaled_err'],
                                                        force_positive=True)

        save_to_cpickle(filename + '_preparation', preparation, config_in['output'], night)

        if defined_reference == night \
                or defined_reference == instrument \
                or defined_reference == 'all' :

                compute_reference = True
                reference_flux[n_night, :] = preparation['coadd']['rescaled']
                reference_mask[n_night, :] = preparation['coadd']['null']
                reference_wght[n_night, :] = 1. / (preparation['coadd']['rescaled_err'] ** 2)

    if compute_reference:
        reference = {
            'wave': shared_data['coadd']['wave'],
            'step': shared_data['coadd']['step'],
            'size': shared_data['coadd']['size'],
        }

        masked_array = np.ma.array(reference_flux, mask=reference_mask)
        rescaled_mask, sum_weights = np.ma.average(masked_array,
                                                   weights=reference_wght,
                                                   axis=0,
                                                   returned=True)

        reference['rescaled'] = rescaled_mask.filled(0.00)
        sum_weights[sum_weights <= 0.0] = 1.0
        reference['rescaled_err'] = 1. / np.sqrt(sum_weights)

        reference['rescaled'], reference['rescaled_err'], reference['null'] = \
            replace_values_errors_with_interpolation_1d(reference['rescaled'],
                                                        reference['rescaled_err'],
                                                        force_positive=True)

        save_to_cpickle(filename + '_reference', reference, config_in['output'])


    print()

